"""
Tests task-purge functionality.
"""
from datetime import datetime, timedelta, timezone

from pulpcore.client.pulpcore import (
    ApiClient,
    ApiException,
    Purge,
    TasksApi,
)
from pulpcore.client.pulp_file import (
    RemotesFileApi,
    RepositoriesFileApi,
    RepositorySyncURL,
)

from pulpcore.constants import TASK_STATES, TASK_FINAL_STATES

from pulp_smash import config
from pulp_smash.pulp3.bindings import (
    monitor_task,
    PulpTestCase,
    PulpTaskError,
)
from .utils import gen_file_remote, gen_repo, gen_file_client, gen_user, del_user

TOMORROW_STR = (datetime.now(timezone.utc) + timedelta(days=1)).strftime("%Y-%m-%dT%H:%M")


class TaskPurgeTestCase(PulpTestCase):
    """
    Test task-purge functionality.
    """

    def _task_summary(self):
        """
        Summary of number of tasks in all known task-states.
        :return: tuple of (total-tasks, Dict(state: count))
        """
        summary = {}
        total = 0
        final_total = 0
        for state in TASK_STATES.__dict__.values():
            response = self.task_api.list(state=state)
            summary[state] = response.count
            total += summary[state]
            final_total += summary[state] if state in TASK_FINAL_STATES else 0
        return total, final_total, summary

    def _purge_report_total(self, task):
        for report in task.progress_reports:
            if report.code == "purge.tasks.total":
                return report.total
        self.fail("NO PURGE_TASKS_TOTAL?!?")

    def _purge_report_check(self, task):
        subobj_total = 0
        total = 0
        for report in task.progress_reports:
            if report["code"] != "purge.tasks.total":
                total = report.total
            else:
                subobj_total += report.total
        self.assertEqual(total, subobj_total)

    def _check_delete_report(self, task, expected):
        # Make sure we reported the deletion
        for report in task.progress_reports:
            if report.code == "purge.tasks.key.core.Task":
                self.assertEqual(report.total, expected)
                break
        else:
            self.fail("NO core.Task DELETIONS?!?")

    @classmethod
    def setUpClass(cls):
        """Create repos, remotes, and api-clients for all tests."""
        cls.cfg = config.get_config()
        cls.client = ApiClient(configuration=cls.cfg.get_bindings_config())
        cls.task_api = TasksApi(cls.client)

        cls.file_client = gen_file_client()
        cls.remote_api = RemotesFileApi(cls.file_client)
        cls.repo_api = RepositoriesFileApi(cls.file_client)

        cls.good_remote = cls.remote_api.create(gen_file_remote(policy="on_demand"))
        cls.good_repo = cls.repo_api.create(gen_repo())
        cls.good_sync_data = RepositorySyncURL(remote=cls.good_remote.pulp_href)

        cls.bad_remote = cls.remote_api.create(
            gen_file_remote(
                "https://fixtures.pulpproject.org/THEREISNOFILEREPOHERE/", policy="on_demand"
            )
        )
        cls.bad_repo = cls.repo_api.create(gen_repo())
        cls.bad_sync_data = RepositorySyncURL(remote=cls.bad_remote.pulp_href)

    @classmethod
    def tearDownClass(cls) -> None:
        """Cleanup repos and remotes. Do the best we can, ignore any errors."""
        try:
            cls.remote_api.delete(cls.bad_remote.pulp_href)
        except:  # noqa
            pass
        try:
            cls.remote_api.delete(cls.good_remote.pulp_href)
        except:  # noqa
            pass
        try:
            cls.repo_api.delete(cls.bad_repo.pulp_href)
        except:  # noqa
            pass
        try:
            cls.repo_api.delete(cls.good_repo.pulp_href)
        except:  # noqa
            pass

    def setUp(self) -> None:
        """
        Give us tasks to operate on, and a summary of tasks before we got here.

        Sets up 1 completed sync, 1 failed.
        """
        self.pre_total, self.pre_final, self.pre_summary = self._task_summary()

        # good sync
        sync_response = self.repo_api.sync(self.good_repo.pulp_href, self.good_sync_data)
        task = monitor_task(sync_response.task)
        self.assertEqual(task.state, "completed")
        self.completed_sync_task = task

        # bad sync
        sync_response = self.repo_api.sync(self.bad_repo.pulp_href, self.bad_sync_data)
        with self.assertRaises(PulpTaskError):
            monitor_task(sync_response.task)
        task = self.task_api.read(sync_response.task)
        self.assertEqual(task.state, "failed")
        self.failed_sync_task = task

        self.post_total, self.post_final, self.post_summary = self._task_summary()
        self.assertEqual(self.post_total, (self.pre_total + 2))
        self.assertEqual(self.post_final, (self.pre_final + 2))

    def test_purge_before_time(self):
        """Purge that should find no tasks to delete."""
        dta = Purge(finished_before="1970-01-01T00:00")
        response = self.task_api.purge(dta)
        task = monitor_task(response.task)
        new_total, new_final, new_summary = self._task_summary()
        # Should have all tasks remaining (2 completed, 1 failed)
        self.assertEqual(self.pre_total + 3, new_total)
        # Should show we report having purged no tasks
        self.assertEqual(self._purge_report_total(task), 0)

    def test_purge_defaults(self):
        """Purge using defaults (finished_before=30-days-ago, state=completed)"""
        dta = Purge()
        response = self.task_api.purge(dta)
        monitor_task(response.task)

        # default is "completed before 30 days ago" - so both sync tasks should still exist
        # Make sure good sync-task still exists
        self.task_api.read(self.completed_sync_task.pulp_href)

        # Make sure the failed sync still exists
        self.task_api.read(self.failed_sync_task.pulp_href)

    def test_purge_all(self):
        """Purge all tasks in any 'final' state."""
        states = list(TASK_FINAL_STATES)
        dta = Purge(finished_before=TOMORROW_STR, states=states)
        response = self.task_api.purge(dta)
        task = monitor_task(response.task)
        new_total, new_final, new_summary = self._task_summary()
        self.assertEqual(1, new_final)  # The purge-task is the only final-task left

        # Make sure good sync-task is gone
        with self.assertRaises(ApiException):
            self.task_api.read(self.completed_sync_task.pulp_href)

        # Make sure failed sync-task is gone
        with self.assertRaises(ApiException):
            self.task_api.read(self.failed_sync_task.pulp_href)

        # Make sure we reported the deletions
        self._check_delete_report(task, self.pre_final + 2)

    def test_purge_leave_one(self):
        """Arrange to leave one task unscathed."""
        # Leave only the failed sync
        dta = Purge(finished_before=self.failed_sync_task.finished_at)
        response = self.task_api.purge(dta)
        task = monitor_task(response.task)

        # Make sure good sync-task is gone
        with self.assertRaises(ApiException):
            self.task_api.read(self.completed_sync_task.pulp_href)

        # Make sure the failed sync still exists
        self.task_api.read(self.failed_sync_task.pulp_href)

        # Make sure we reported the task-deletion
        self._check_delete_report(task, self.pre_summary["completed"] + 1)

    def test_purge_only_failed(self):
        """Purge all failed tasks only."""
        dta = Purge(finished_before=TOMORROW_STR, states=["failed"])
        response = self.task_api.purge(dta)
        monitor_task(response.task)
        # completed sync-task should exist
        self.task_api.read(self.completed_sync_task.pulp_href)

        # failed should not exist
        with self.assertRaises(ApiException):
            self.task_api.read(self.failed_sync_task.pulp_href)

    def test_bad_date(self):
        """What happens if you use a bad date format?"""
        dta = Purge(finished_before="THISISNOTADATE")
        with self.assertRaises(ApiException):
            self.task_api.purge(dta)

    def test_bad_state(self):
        """What happens if you specify junk for a state?"""
        dta = Purge(finished_before=TOMORROW_STR, states=["BAD STATE"])
        with self.assertRaises(ApiException):
            self.task_api.purge(dta)

    def test_not_final_state(self):
        """What happens if you use a valid state that isn't a 'final' one?"""
        dta = Purge(finished_before=TOMORROW_STR, states=["running"])
        with self.assertRaises(ApiException):
            self.task_api.purge(dta)


class TaskPurgeUserPermsTestCase(PulpTestCase):
    """
    Test task-purge is correctly protected by user-perms.

    Create new-user
    Sync as admin
    Purge as new-user, sync-task should NOT be deleted
    """

    @classmethod
    def setUpClass(cls):
        """Create repos, remotes, and api-clients for all tests."""
        cls.cfg = config.get_config()
        cls.client = ApiClient(configuration=cls.cfg.get_bindings_config())
        file_client = gen_file_client()
        cls.admin_info = {
            "task_api": TasksApi(cls.client),
            "file_client": file_client,
            "remote_api": RemotesFileApi(file_client),
            "repo_api": RepositoriesFileApi(file_client),
        }
        cls.admin_info["a_remote"] = cls.admin_info["remote_api"].create(
            gen_file_remote(policy="on_demand")
        )
        cls.admin_info["a_repo"] = cls.admin_info["repo_api"].create(gen_repo())
        cls.admin_info["sync_data"] = RepositorySyncURL(remote=cls.admin_info["a_remote"].pulp_href)

        cls.new_user = gen_user()
        file_client = gen_file_client()
        cls.user_info = {
            "task_api": TasksApi(cls.client),
            "file_client": file_client,
            "remote_api": RemotesFileApi(file_client),
            "repo_api": RepositoriesFileApi(file_client),
        }
        cls.user_info["a_remote"] = cls.user_info["remote_api"].create(
            gen_file_remote(policy="on_demand")
        )
        cls.user_info["a_repo"] = cls.user_info["repo_api"].create(gen_repo())
        cls.user_info["sync_data"] = RepositorySyncURL(remote=cls.user_info["a_remote"].pulp_href)

    @classmethod
    def tearDownClass(cls) -> None:
        """Cleanup repos and remotes."""
        cls.admin_info["remote_api"].delete(cls.admin_info["a_remote"].pulp_href)
        cls.admin_info["repo_api"].delete(cls.admin_info["a_repo"].pulp_href)
        cls.user_info["remote_api"].delete(cls.user_info["a_remote"].pulp_href)
        cls.user_info["repo_api"].delete(cls.user_info["a_repo"].pulp_href)
        del_user(cls.new_user)

    def testUserCannotPurge(self) -> None:
        """
        Test that purge does NOT purge tasks NOT OWNED by caller.
        """
        # Sync as admin
        sync_response = self.admin_info["repo_api"].sync(
            self.admin_info["a_repo"].pulp_href, self.admin_info["sync_data"]
        )
        task = monitor_task(sync_response.task)
        self.assertEqual(task.state, "completed")

        # Purge as user
        states = list(TASK_FINAL_STATES)
        dta = Purge(finished_before=TOMORROW_STR, states=states)
        response = self.user_info["task_api"].purge(dta)
        task = monitor_task(response.task)

        # Make sure sync-task (executed by admin) still exists
        self.admin_info["task_api"].read(task.pulp_href)

    def testUserCanPurge(self) -> None:
        """
        Test that purge DOES purge tasks owned by caller.
        """
        # Sync as user
        sync_response = self.user_info["repo_api"].sync(
            self.user_info["a_repo"].pulp_href, self.user_info["sync_data"]
        )
        sync_task = monitor_task(sync_response.task)
        self.assertEqual(sync_task.state, "completed")

        # Purge as user
        states = list(TASK_FINAL_STATES)
        dta = Purge(finished_before=TOMORROW_STR, states=states)
        response = self.user_info["task_api"].purge(dta)
        monitor_task(response.task)

        # Make sure task DOES NOT exist
        with self.assertRaises(ApiException):
            self.admin_info["task_api"].read(sync_task.pulp_href)

    def testAdminCanPurge(self) -> None:
        """
        Test that admin can ALWAYS purge.
        """
        # Sync as user
        sync_response = self.user_info["repo_api"].sync(
            self.user_info["a_repo"].pulp_href, self.user_info["sync_data"]
        )
        sync_task = monitor_task(sync_response.task)
        self.assertEqual(sync_task.state, "completed")

        # Purge as ADMIN
        states = list(TASK_FINAL_STATES)
        dta = Purge(finished_before=TOMORROW_STR, states=states)
        response = self.admin_info["task_api"].purge(dta)
        monitor_task(response.task)

        # Make sure task DOES NOT exist
        with self.assertRaises(ApiException):
            self.admin_info["task_api"].read(sync_task.pulp_href)



# Generated by Django 3.2.6 on 2021-08-30 11:27

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0075_rbaccontentguard'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='taskreservedresource',
            name='resource',
        ),
        migrations.RemoveField(
            model_name='taskreservedresource',
            name='task',
        ),
        migrations.RemoveField(
            model_name='worker',
            name='cleaned_up',
        ),
        migrations.RemoveField(
            model_name='worker',
            name='gracefully_stopped',
        ),
        migrations.RemoveField(
            model_name='task',
            name='_resource_job_id',
        ),
        migrations.DeleteModel(
            name='ReservedResource',
        ),
        migrations.DeleteModel(
            name='TaskReservedResource',
        ),
    ]



"""Tests that perform actions over distributions."""
import csv
import hashlib
from time import sleep
import unittest
from urllib.parse import urljoin

from pulp_smash import api, cli, config, utils
from pulp_smash.pulp3.bindings import delete_orphans
from pulp_smash.pulp3.utils import (
    download_content_unit,
    download_content_unit_return_requests_response,
    gen_distribution,
    gen_repo,
    get_content,
    get_versions,
    modify_repo,
    sync,
    utils as pulp3_utils,
)
from requests.exceptions import HTTPError

from pulpcore.client.pulpcore import ApiException, StatusApi

from pulpcore.client.pulp_file import (
    ContentFilesApi,
    DistributionsFileApi,
    FileFilePublication,
    PublicationsFileApi,
    RemotesFileApi,
    RepositoriesFileApi,
    RepositorySyncURL,
)
from pulpcore.tests.functional.api.using_plugin.constants import (
    FILE_CHUNKED_FIXTURE_MANIFEST_URL,
    FILE_CONTENT_NAME,
    FILE_DISTRIBUTION_PATH,
    FILE_FIXTURE_COUNT,
    FILE_REMOTE_PATH,
    FILE_URL,
    FILE_REPO_PATH,
)
from pulpcore.tests.functional.api.using_plugin.utils import (
    create_file_publication,
    gen_file_remote,
    gen_file_client,
    gen_pulpcore_client,
    monitor_task,
)
from pulpcore.tests.functional.api.using_plugin.utils import set_up_module as setUpModule  # noqa
from pulpcore.tests.functional.api.using_plugin.utils import skip_if


class CRUDPublicationDistributionTestCase(unittest.TestCase):
    """CRUD Publication Distribution.

    This test targets the following issue:

    * `Pulp #4839 <https://pulp.plan.io/issues/4839>`_
    * `Pulp #4862 <https://pulp.plan.io/issues/4862>`_
    """

    @classmethod
    def setUpClass(cls):
        """Create class-wide variables."""
        cls.cfg = config.get_config()
        cls.client = api.Client(cls.cfg)
        cls.attr = (
            "name",
            "base_path",
        )
        cls.distribution = {}
        cls.publication = {}
        cls.remote = {}
        cls.repo = {}

    @classmethod
    def tearDownClass(cls):
        """Clean class-wide variables."""
        for resource in (cls.publication, cls.remote, cls.repo):
            if resource:
                cls.client.delete(resource["pulp_href"])

    def test_01_create(self):
        """Create a publication distribution.

        Do the following:

        1. Create a repository and 3 repository versions with at least 1 file
           content in it. Create a publication using the second repository
           version.
        2. Create a distribution with 'publication' field set to
           the publication from step (1).
        3. Assert the distribution got created correctly with the correct
           base_path, name, and publication. Assert that content guard is
           unset.
        4. Assert that publication has a 'distributions' reference to the
           distribution (it's backref).

        """
        self.repo.update(self.client.post(FILE_REPO_PATH, gen_repo()))
        self.remote.update(self.client.post(FILE_REMOTE_PATH, gen_file_remote()))
        # create 3 repository versions
        sync(self.cfg, self.remote, self.repo)
        self.repo = self.client.get(self.repo["pulp_href"])
        for file_content in get_content(self.repo)[FILE_CONTENT_NAME]:
            modify_repo(self.cfg, self.repo, remove_units=[file_content])

        self.repo = self.client.get(self.repo["pulp_href"])

        versions = get_versions(self.repo)

        self.publication.update(
            create_file_publication(self.cfg, self.repo, versions[1]["pulp_href"])
        )

        self.distribution.update(
            self.client.post(
                FILE_DISTRIBUTION_PATH, gen_distribution(publication=self.publication["pulp_href"])
            )
        )

        self.publication = self.client.get(self.publication["pulp_href"])

        # content_guard and repository parameters unset.
        for key, val in self.distribution.items():
            if key in ["content_guard", "repository"]:
                self.assertIsNone(val, self.distribution)
            else:
                self.assertIsNotNone(val, self.distribution)

        self.assertEqual(
            self.distribution["publication"], self.publication["pulp_href"], self.distribution
        )

        self.assertEqual(
            self.publication["distributions"][0], self.distribution["pulp_href"], self.publication
        )

    @skip_if(bool, "distribution", False)
    def test_02_read(self):
        """Read distribution by its href."""
        distribution = self.client.get(self.distribution["pulp_href"])
        for key, val in self.distribution.items():
            with self.subTest(key=key):
                self.assertEqual(distribution[key], val)

    @skip_if(bool, "distribution", False)
    def test_03_partially_update(self):
        """Update a distribution using PATCH."""
        for key in self.attr:
            with self.subTest(key=key):
                self.do_partially_update_attr(key)

    @skip_if(bool, "distribution", False)
    def test_03_fully_update(self):
        """Update a distribution using PUT."""
        for key in self.attr:
            with self.subTest(key=key):
                self.do_fully_update_attr(key)

    @skip_if(bool, "distribution", False)
    def test_04_delete_distribution(self):
        """Delete a distribution."""
        self.client.delete(self.distribution["pulp_href"])
        with self.assertRaises(HTTPError):
            self.client.get(self.distribution["pulp_href"])

    def do_fully_update_attr(self, attr):
        """Update a distribution attribute using HTTP PUT.

        :param attr: The name of the attribute to update.
        """
        distribution = self.client.get(self.distribution["pulp_href"])
        string = utils.uuid4()
        distribution[attr] = string
        self.client.put(distribution["pulp_href"], distribution)

        # verify the update
        distribution = self.client.get(distribution["pulp_href"])
        self.assertEqual(string, distribution[attr], distribution)

    def do_partially_update_attr(self, attr):
        """Update a distribution using HTTP PATCH.

        :param attr: The name of the attribute to update.
        """
        string = utils.uuid4()
        self.client.patch(self.distribution["pulp_href"], {attr: string})

        # Verify the update
        distribution = self.client.get(self.distribution["pulp_href"])
        self.assertEqual(string, distribution[attr], self.distribution)


class DistributionBasePathTestCase(unittest.TestCase):
    """Test possible values for ``base_path`` on a distribution.

    This test targets the following issues:

    * `Pulp #2987 <https://pulp.plan.io/issues/2987>`_
    * `Pulp #3412 <https://pulp.plan.io/issues/3412>`_
    * `Pulp #4882 <https://pulp.plan.io/issues/4882>`_
    * `Pulp Smash #906 <https://github.com/pulp/pulp-smash/issues/906>`_
    * `Pulp Smash #956 <https://github.com/pulp/pulp-smash/issues/956>`_
    """

    @classmethod
    def setUpClass(cls):
        """Create class-wide variables."""
        cls.cfg = config.get_config()
        cls.client = api.Client(cls.cfg)
        body = gen_distribution()
        body["base_path"] = body["base_path"].replace("-", "/")
        distribution = cls.client.post(FILE_DISTRIBUTION_PATH, body)
        cls.distribution = cls.client.get(distribution["pulp_href"])

    @classmethod
    def tearDownClass(cls):
        """Clean up resources."""
        cls.client.delete(cls.distribution["pulp_href"])

    def test_negative_create_using_spaces(self):
        """Test that spaces can not be part of ``base_path``."""
        self.try_create_distribution(base_path=utils.uuid4().replace("-", " "))
        self.try_update_distribution(base_path=utils.uuid4().replace("-", " "))

    def test_negative_create_using_begin_slash(self):
        """Test that slash cannot be in the begin of ``base_path``."""
        self.try_create_distribution(base_path="/" + utils.uuid4())
        self.try_update_distribution(base_path="/" + utils.uuid4())

    def test_negative_create_using_end_slash(self):
        """Test that slash cannot be in the end of ``base_path``."""
        self.try_create_distribution(base_path=utils.uuid4() + "/")
        self.try_update_distribution(base_path=utils.uuid4() + "/")

    def test_negative_create_using_non_unique_base_path(self):
        """Test that ``base_path`` can not be duplicated."""
        self.try_create_distribution(base_path=self.distribution["base_path"])

    def test_negative_create_using_overlapping_base_path(self):
        """Test that distributions can't have overlapping ``base_path``.

        See: `Pulp #2987`_.
        """
        base_path = self.distribution["base_path"].rsplit("/", 1)[0]
        self.try_create_distribution(base_path=base_path)

        base_path = "/".join((self.distribution["base_path"], utils.uuid4().replace("-", "/")))
        self.try_create_distribution(base_path=base_path)

    def try_create_distribution(self, **kwargs):
        """Unsuccessfully create a distribution.

        Merge the given kwargs into the body of the request.
        """
        body = gen_distribution()
        body.update(kwargs)
        with self.assertRaises(HTTPError) as ctx:
            self.client.post(FILE_DISTRIBUTION_PATH, body)

        self.assertIsNotNone(
            ctx.exception.response.json()["base_path"], ctx.exception.response.json()
        )

    def try_update_distribution(self, **kwargs):
        """Unsuccessfully update a distribution with HTTP PATCH.

        Use the given kwargs as the body of the request.
        """
        with self.assertRaises(HTTPError) as ctx:
            self.client.patch(self.distribution["pulp_href"], kwargs)

        self.assertIsNotNone(
            ctx.exception.response.json()["base_path"], ctx.exception.response.json()
        )


class ContentServePublicationDistributionTestCase(unittest.TestCase):
    """Verify that content is served from a publication distribution.

    Assert that published metadata and content is served from a publication
    distribution.

    This test targets the following issue:

    `Pulp #4847 <https://pulp.plan.io/issues/4847>`_
    """

    @classmethod
    def setUpClass(cls):
        """Create class-wide variables."""
        cls.cfg = config.get_config()
        cls.client = gen_file_client()

        cls.content_api = ContentFilesApi(cls.client)
        cls.repo_api = RepositoriesFileApi(cls.client)
        cls.remote_api = RemotesFileApi(cls.client)
        cls.publications_api = PublicationsFileApi(cls.client)
        cls.distributions_api = DistributionsFileApi(cls.client)

    def setUp(self):
        delete_orphans()

    def test_nonpublished_content_not_served(self):
        """Verify content that hasn't been published is not served."""
        self.setup_download_test("immediate", publish=False)
        files = ["", "1.iso", "2.iso", "3.iso"]
        for file in files:
            with self.assertRaises(HTTPError, msg=f"{file}") as cm:
                download_content_unit(self.cfg, self.distribution.to_dict(), file)
            self.assertEqual(cm.exception.response.status_code, 404, f"{file}")

    def test_content_served_on_demand(self):
        """Assert that on_demand content can be properly downloaded."""
        self.setup_download_test("on_demand")
        self.do_test_content_served()

    def test_content_served_immediate(self):
        """Assert that downloaded content can be properly downloaded."""
        self.setup_download_test("immediate")
        self.do_test_content_served()

    def test_content_served_streamed(self):
        """Assert that streamed content can be properly downloaded."""
        self.setup_download_test("streamed")
        self.do_test_content_served()

    def test_content_served_immediate_with_range_request_inside_one_chunk(self):
        """Assert that downloaded content can be properly downloaded with range requests."""
        self.setup_download_test("immediate", url=FILE_CHUNKED_FIXTURE_MANIFEST_URL)
        range_headers = {"Range": "bytes=1048586-1049586"}
        num_bytes = 1001
        self.do_range_request_download_test(range_headers, num_bytes)

    def test_content_served_immediate_with_range_request_over_three_chunks(self):
        """Assert that downloaded content can be properly downloaded with range requests."""
        self.setup_download_test("immediate", url=FILE_CHUNKED_FIXTURE_MANIFEST_URL)
        range_headers = {"Range": "bytes=1048176-2248576"}
        num_bytes = 1200401
        self.do_range_request_download_test(range_headers, num_bytes)

    def test_content_served_on_demand_with_range_request_over_three_chunks(self):
        """Assert that on_demand content can be properly downloaded with range requests."""
        self.setup_download_test("on_demand", url=FILE_CHUNKED_FIXTURE_MANIFEST_URL)
        range_headers = {"Range": "bytes=1048176-2248576"}
        num_bytes = 1200401
        self.do_range_request_download_test(range_headers, num_bytes)

    def test_content_served_streamed_with_range_request_over_three_chunks(self):
        """Assert that streamed content can be properly downloaded with range requests."""
        self.setup_download_test("streamed", url=FILE_CHUNKED_FIXTURE_MANIFEST_URL)
        range_headers = {"Range": "bytes=1048176-2248576"}
        num_bytes = 1200401
        self.do_range_request_download_test(range_headers, num_bytes)

    def test_content_served_immediate_with_multiple_different_range_requests(self):
        """Assert that multiple requests with different Range header values work as expected."""
        self.setup_download_test("immediate", url=FILE_CHUNKED_FIXTURE_MANIFEST_URL)
        range_headers = {"Range": "bytes=1048176-2248576"}
        num_bytes = 1200401
        self.do_range_request_download_test(range_headers, num_bytes)
        range_headers = {"Range": "bytes=2042176-3248576"}
        num_bytes = 1206401
        self.do_range_request_download_test(range_headers, num_bytes)

    def test_content_served_immediate_with_range_request_invalid_start_value(self):
        """Assert that range requests with a negative start value errors as expected."""
        cfg = config.get_config()
        cli_client = cli.Client(cfg)
        storage = utils.get_pulp_setting(cli_client, "DEFAULT_FILE_STORAGE")
        if storage != "pulpcore.app.models.storage.FileSystem":
            self.skipTest("The S3 test API project doesn't handle invalid Range values correctly")
        self.setup_download_test("immediate", url=FILE_CHUNKED_FIXTURE_MANIFEST_URL)
        with self.assertRaises(HTTPError) as cm:
            download_content_unit_return_requests_response(
                self.cfg, self.distribution.to_dict(), "1.iso", headers={"Range": "bytes=-1-11"}
            )
        self.assertEqual(cm.exception.response.status_code, 416)

    def test_content_served_immediate_with_range_request_too_large_end_value(self):
        """Assert that a range request with a end value that is larger than the data works still."""
        self.setup_download_test("immediate", url=FILE_CHUNKED_FIXTURE_MANIFEST_URL)
        range_headers = {"Range": "bytes=10485260-10485960"}
        num_bytes = 500
        self.do_range_request_download_test(range_headers, num_bytes)

    def test_content_served_immediate_with_range_request_start_value_larger_than_content(self):
        """Assert that a range request with a start value larger than the content errors."""
        self.setup_download_test("immediate", url=FILE_CHUNKED_FIXTURE_MANIFEST_URL)
        with self.assertRaises(HTTPError) as cm:
            download_content_unit_return_requests_response(
                self.cfg,
                self.distribution.to_dict(),
                "1.iso",
                headers={"Range": "bytes=10485860-10485870"},
            )
        self.assertEqual(cm.exception.response.status_code, 416)

    def test_content_served_after_db_restart(self):
        """
        Assert that content can be downloaded after the database has been restarted.
        This test also check that the HTML page with a list of distributions is also
        available after the connection to the database has been closed.
        """
        cfg = config.get_config()
        pulp_host = cfg.hosts[0]
        svc_mgr = cli.ServiceManager(cfg, pulp_host)
        if svc_mgr._svc_mgr == "s6":
            postgresql_service_name = "postgresql"
        else:
            postgresql_service_name = "*postgresql*"
        postgresql_found = svc_mgr.is_active([postgresql_service_name])
        self.assertTrue(
            postgresql_found, "PostgreSQL service not found or is not active. Can't restart it."
        )
        svc_mgr.restart([postgresql_service_name])
        # Wait for postgres to come back and pulpcore-api to recover
        status_api = StatusApi(gen_pulpcore_client())
        for i in range(5):
            sleep(2)
            try:
                status_api.status_read()
                break
            except ApiException:
                if i == 4:
                    raise
        self.setup_download_test("immediate")
        self.do_test_content_served()
        url_fragments = [
            cfg.get_content_host_base_url(),
            "pulp/content",
        ]
        content_app_root = "/".join(url_fragments)
        pulp3_utils.http_get(content_app_root)

    def setup_download_test(self, policy, url=None, publish=True):
        # Create a repository
        self.repo = self.repo_api.create(gen_repo())
        self.addCleanup(self.repo_api.delete, self.repo.pulp_href)

        # Create a remote
        remote_options = {"policy": policy}
        if url:
            remote_options["url"] = url

        self.remote = self.remote_api.create(gen_file_remote(**remote_options))
        self.addCleanup(self.remote_api.delete, self.remote.pulp_href)

        # Sync the repository.
        repository_sync_data = RepositorySyncURL(remote=self.remote.pulp_href)
        sync_response = self.repo_api.sync(self.repo.pulp_href, repository_sync_data)
        monitor_task(sync_response.task)

        if publish:
            # Create a publication.
            publish_data = FileFilePublication(repository=self.repo.pulp_href)
            publish_response = self.publications_api.create(publish_data)
            publication_href = monitor_task(publish_response.task).created_resources[0]
            self.addCleanup(self.publications_api.delete, publication_href)
            serve, served_href = "publication", publication_href
        else:
            serve, served_href = "repository", self.repo.pulp_href

        # Create a distribution.
        response = self.distributions_api.create(gen_distribution(**{serve: served_href}))
        distribution_href = monitor_task(response.task).created_resources[0]
        self.distribution = self.distributions_api.read(distribution_href)
        self.addCleanup(self.distributions_api.delete, self.distribution.pulp_href)

    def do_test_content_served(self):
        file_path = "1.iso"

        req1 = download_content_unit(self.cfg, self.distribution.to_dict(), file_path)
        req2 = download_content_unit(self.cfg, self.distribution.to_dict(), file_path)
        fixtures_hash = hashlib.sha256(utils.http_get(urljoin(FILE_URL, file_path))).hexdigest()

        first_dl_hash = hashlib.sha256(req1).hexdigest()
        second_dl_hash = hashlib.sha256(req2).hexdigest()

        self.assertEqual(first_dl_hash, fixtures_hash)
        self.assertEqual(first_dl_hash, second_dl_hash)

        manifest = download_content_unit(self.cfg, self.distribution.to_dict(), "PULP_MANIFEST")
        pulp_manifest = list(
            csv.DictReader(manifest.decode("utf-8").splitlines(), ("name", "checksum", "size"))
        )

        self.assertEqual(len(pulp_manifest), FILE_FIXTURE_COUNT, pulp_manifest)

    def do_range_request_download_test(self, range_header, expected_bytes):
        file_path = "1.iso"

        req1_reponse = download_content_unit_return_requests_response(
            self.cfg, self.distribution.to_dict(), file_path, headers=range_header
        )
        req2_response = download_content_unit_return_requests_response(
            self.cfg, self.distribution.to_dict(), file_path, headers=range_header
        )

        self.assertEqual(expected_bytes, len(req1_reponse.content))
        self.assertEqual(expected_bytes, len(req2_response.content))
        self.assertEqual(req1_reponse.content, req2_response.content)

        self.assertEqual(req1_reponse.status_code, 206)
        self.assertEqual(req1_reponse.status_code, req2_response.status_code)

        self.assertEqual(str(expected_bytes), req1_reponse.headers["Content-Length"])
        self.assertEqual(str(expected_bytes), req2_response.headers["Content-Length"])

        self.assertEqual(
            req1_reponse.headers["Content-Range"], req2_response.headers["Content-Range"]
        )



import asyncio
from collections import defaultdict
from gettext import gettext as _
import logging

from asgiref.sync import sync_to_async
from django.db.models import Prefetch, prefetch_related_objects, Q

from pulpcore.plugin.exceptions import UnsupportedDigestValidationError
from pulpcore.plugin.models import (
    AlternateContentSource,
    Artifact,
    ContentArtifact,
    ProgressReport,
    RemoteArtifact,
)
from pulpcore.plugin.sync import sync_to_async_iterable

from .api import Stage

log = logging.getLogger(__name__)


def _check_for_forbidden_checksum_type(artifact):
    """Check if content doesn't have forbidden checksum type.

    If contains forbidden checksum type it will raise ValueError,
    otherwise it passes without returning anything.
    """
    for digest_type in Artifact.FORBIDDEN_DIGESTS:
        digest_value = getattr(artifact, digest_type)
        if digest_value:
            # To use shared message constant when #7988 is merged
            raise UnsupportedDigestValidationError(
                _(
                    "Artifact contains forbidden checksum type {}. You can allow it with "
                    "'ALLOWED_CONTENT_CHECKSUMS' setting."
                ).format(digest_type)
            )


class QueryExistingArtifacts(Stage):
    """
    A Stages API stage that replaces :attr:`DeclarativeContent.content` objects with already-saved
    :class:`~pulpcore.plugin.models.Artifact` objects.

    This stage expects :class:`~pulpcore.plugin.stages.DeclarativeContent` units from `self._in_q`
    and inspects their associated :class:`~pulpcore.plugin.stages.DeclarativeArtifact` objects. Each
    :class:`~pulpcore.plugin.stages.DeclarativeArtifact` object stores one
    :class:`~pulpcore.plugin.models.Artifact`.

    This stage inspects any unsaved :class:`~pulpcore.plugin.models.Artifact` objects and searches
    using their metadata for existing saved :class:`~pulpcore.plugin.models.Artifact` objects inside
    Pulp with the same digest value(s). Any existing :class:`~pulpcore.plugin.models.Artifact`
    objects found will replace their unsaved counterpart in the
    :class:`~pulpcore.plugin.stages.DeclarativeArtifact` object.

    Each :class:`~pulpcore.plugin.stages.DeclarativeContent` is sent to `self._out_q` after all of
    its :class:`~pulpcore.plugin.stages.DeclarativeArtifact` objects have been handled.

    This stage drains all available items from `self._in_q` and batches everything into one large
    call to the db for efficiency.
    """

    async def run(self):
        """
        The coroutine for this stage.

        Returns:
            The coroutine for this stage.
        """
        async for batch in self.batches():
            artifact_digests_by_type = defaultdict(list)

            # For each unsaved artifact, check its digests in the order of COMMON_DIGEST_FIELDS
            # and the first digest which is found is added to the list of digests of that type.
            # We assume that in general only one digest is provided and that it will be
            # sufficient to identify the Artifact.
            for d_content in batch:
                for d_artifact in d_content.d_artifacts:
                    if d_artifact.artifact._state.adding:
                        if not d_artifact.deferred_download:
                            _check_for_forbidden_checksum_type(d_artifact.artifact)
                        for digest_type in Artifact.COMMON_DIGEST_FIELDS:
                            digest_value = getattr(d_artifact.artifact, digest_type)
                            if digest_value:
                                artifact_digests_by_type[digest_type].append(digest_value)
                                break

            # For each type of digest, fetch all the existing Artifacts where digest "in"
            # the list we built earlier. Walk over all the artifacts again compare the
            # digest of the new artifact to those of the existing ones - if one matches,
            # swap it out with the existing one.
            for digest_type, digests in artifact_digests_by_type.items():
                query_params = {"{attr}__in".format(attr=digest_type): digests}
                existing_artifacts_qs = Artifact.objects.filter(**query_params)
                existing_artifacts = sync_to_async_iterable(existing_artifacts_qs)
                await sync_to_async(existing_artifacts_qs.touch)()
                for d_content in batch:
                    for d_artifact in d_content.d_artifacts:
                        artifact_digest = getattr(d_artifact.artifact, digest_type)
                        if artifact_digest:
                            async for result in existing_artifacts:
                                result_digest = getattr(result, digest_type)
                                if result_digest == artifact_digest:
                                    d_artifact.artifact = result
                                    break
            for d_content in batch:
                await self.put(d_content)


class ArtifactDownloader(Stage):
    """
    A Stages API stage to download :class:`~pulpcore.plugin.models.Artifact` files, but don't save
    the :class:`~pulpcore.plugin.models.Artifact` in the db.

    This stage downloads the file for any :class:`~pulpcore.plugin.models.Artifact` objects missing
    files and creates a new :class:`~pulpcore.plugin.models.Artifact` object from the downloaded
    file and its digest data. The new :class:`~pulpcore.plugin.models.Artifact` is not saved but
    added to the :class:`~pulpcore.plugin.stages.DeclarativeArtifact` object, replacing the likely
    incomplete :class:`~pulpcore.plugin.models.Artifact`.

    Each :class:`~pulpcore.plugin.stages.DeclarativeContent` is sent to `self._out_q` after all of
    its :class:`~pulpcore.plugin.stages.DeclarativeArtifact` objects have been handled.

    This stage creates a ProgressReport named 'Downloading Artifacts' that counts the number of
    downloads completed. Since it's a stream the total count isn't known until it's finished.

    This stage drains all available items from `self._in_q` and starts as many downloaders as
    possible (up to `download_concurrency` set on a Remote)

    Args:
        max_concurrent_content (int): The maximum number of
            :class:`~pulpcore.plugin.stages.DeclarativeContent` instances to handle simultaneously.
            Default is 200.
        args: unused positional arguments passed along to :class:`~pulpcore.plugin.stages.Stage`.
        kwargs: unused keyword arguments passed along to :class:`~pulpcore.plugin.stages.Stage`.
    """

    def __init__(self, max_concurrent_content=200, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.max_concurrent_content = max_concurrent_content

    async def run(self):
        """
        The coroutine for this stage.

        Returns:
            The coroutine for this stage.
        """

        def _add_to_pending(coro):
            nonlocal pending
            task = asyncio.ensure_future(coro)
            pending.add(task)
            return task

        #: (set): The set of unfinished tasks.  Contains the content
        #    handler tasks and may contain `content_get_task`.
        pending = set()

        content_iterator = self.items()

        #: (:class:`asyncio.Task`): The task that gets new content from `self._in_q`.
        #    Set to None if stage is shutdown.
        content_get_task = _add_to_pending(content_iterator.__anext__())

        async with ProgressReport(
            message="Downloading Artifacts", code="sync.downloading.artifacts"
        ) as pb:
            try:
                while pending:
                    done, pending = await asyncio.wait(pending, return_when=asyncio.FIRST_COMPLETED)
                    for task in done:
                        if task is content_get_task:
                            try:
                                _add_to_pending(self._handle_content_unit(task.result()))
                            except StopAsyncIteration:
                                # previous stage is finished and we retrieved all
                                # content instances: shutdown
                                content_get_task = None
                        else:
                            pb.done += task.result()  # download_count
                            await pb.asave()

                    if content_get_task and content_get_task not in pending:  # not yet shutdown
                        if len(pending) < self.max_concurrent_content:
                            content_get_task = _add_to_pending(content_iterator.__anext__())
            except asyncio.CancelledError:
                # asyncio.wait does not cancel its tasks when cancelled, we need to do this
                for future in pending:
                    future.cancel()
                raise

    async def _handle_content_unit(self, d_content):
        """Handle one content unit.

        Returns:
            The number of downloads
        """
        downloaders_for_content = [
            d_artifact.download()
            for d_artifact in d_content.d_artifacts
            if d_artifact.artifact._state.adding
            and not d_artifact.deferred_download
            and not d_artifact.artifact.file
        ]
        if downloaders_for_content:
            await asyncio.gather(*downloaders_for_content)
        await self.put(d_content)
        return len(downloaders_for_content)


class ArtifactSaver(Stage):
    """
    A Stages API stage that saves any unsaved :attr:`DeclarativeArtifact.artifact` objects.

    This stage expects :class:`~pulpcore.plugin.stages.DeclarativeContent` units from `self._in_q`
    and inspects their associated :class:`~pulpcore.plugin.stages.DeclarativeArtifact` objects. Each
    :class:`~pulpcore.plugin.stages.DeclarativeArtifact` object stores one
    :class:`~pulpcore.plugin.models.Artifact`.

    Any unsaved :class:`~pulpcore.plugin.models.Artifact` objects are saved. Each
    :class:`~pulpcore.plugin.stages.DeclarativeContent` is sent to `self._out_q` after all of its
    :class:`~pulpcore.plugin.stages.DeclarativeArtifact` objects have been handled.

    This stage drains all available items from `self._in_q` and batches everything into one large
    call to the db for efficiency.
    """

    async def run(self):
        """
        The coroutine for this stage.

        Returns:
            The coroutine for this stage.
        """
        async for batch in self.batches():
            da_to_save = []
            for d_content in batch:
                for d_artifact in d_content.d_artifacts:
                    if d_artifact.artifact._state.adding and not d_artifact.deferred_download:
                        d_artifact.artifact.file = str(d_artifact.artifact.file)
                        da_to_save.append(d_artifact)
            da_to_save_ordered = sorted(da_to_save, key=lambda x: x.artifact.sha256)

            if da_to_save:
                for d_artifact, artifact in zip(
                    da_to_save_ordered,
                    await sync_to_async(Artifact.objects.bulk_get_or_create)(
                        d_artifact.artifact for d_artifact in da_to_save_ordered
                    ),
                ):
                    d_artifact.artifact = artifact

            for d_content in batch:
                await self.put(d_content)


class RemoteArtifactSaver(Stage):
    """
    A Stage that saves :class:`~pulpcore.plugin.models.RemoteArtifact` objects

    An :class:`~pulpcore.plugin.models.RemoteArtifact` object is saved for each
    :class:`~pulpcore.plugin.stages.DeclarativeArtifact`.
    """

    def __init__(self, fix_mismatched_remote_artifacts=False, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fix_mismatched_remote_artifacts = fix_mismatched_remote_artifacts

    async def run(self):
        """
        The coroutine for this stage.

        Returns:
            The coroutine for this stage.
        """
        async for batch in self.batches():
            await self._handle_remote_artifacts(batch)
            for d_content in batch:
                await self.put(d_content)

    async def _handle_remote_artifacts(self, batch):
        """
        Build a list of only :class:`~pulpcore.plugin.models.RemoteArtifact` that need
        to be created for the batch.

        Args:
            batch (list): List of :class:`~pulpcore.plugin.stages.DeclarativeContent`.

        Returns:
            List: Of :class:`~pulpcore.plugin.models.RemoteArtifact`.
        """
        remotes_present = set()
        for d_content in batch:
            # https://code.djangoproject.com/ticket/33596
            # If the content was pre-fetched previously, remove that cached data, which could be out
            # of date.
            if hasattr(d_content.content, "_remote_artifact_saver_cas"):
                delattr(d_content.content, "_remote_artifact_saver_cas")
            for d_artifact in d_content.d_artifacts:
                if d_artifact.remote:
                    remotes_present.add(d_artifact.remote)

        await sync_to_async(prefetch_related_objects)(
            [d_c.content for d_c in batch],
            Prefetch(
                "contentartifact_set",
                queryset=ContentArtifact.objects.prefetch_related(
                    Prefetch(
                        "remoteartifact_set",
                        queryset=RemoteArtifact.objects.filter(remote__in=remotes_present),
                        to_attr="_remote_artifact_saver_ras",
                    )
                ),
                to_attr="_remote_artifact_saver_cas",
            ),
        )

        # Now return the list of RemoteArtifacts that need to be saved.
        #
        # We can end up with duplicates (diff pks, same sha256) in the sequence below,
        # so we store by-sha256 and then return the final values
        ras_to_create = {}  # { str(<sha256>): RemoteArtifact, ... }
        ras_to_update = {}
        for d_content in batch:
            for d_artifact in d_content.d_artifacts:
                if not d_artifact.remote:
                    continue

                async for content_artifact in sync_to_async_iterable(
                    d_content.content._remote_artifact_saver_cas
                ):
                    if d_artifact.relative_path == content_artifact.relative_path:
                        break
                else:
                    if self.fix_mismatched_remote_artifacts:
                        # We couldn't match an DeclarativeArtifact to a ContentArtifact by rel_path.
                        # If there are any paths available (i.e., other ContentArtifacts for this
                        # Artifact), complain to the logs, pick the rel_path from the last
                        # ContentArtifact we examined, and continue.
                        #
                        # If we can't find anything to choose from (can that even happen?), fail
                        # the process.
                        avail_paths = ",".join(
                            [
                                ca.relative_path
                                for ca in d_content.content._remote_artifact_saver_cas
                            ]
                        )
                        if avail_paths:
                            msg = _(
                                "No declared artifact with relative path '{rp}' for content '{c}'"
                                " from remote '{rname}'. Using last from available-paths : '{ap}'"
                            )
                            log.warning(
                                msg.format(
                                    rp=d_artifact.relative_path,
                                    c=d_content.content.natural_key(),
                                    rname=d_artifact.remote.name,
                                    ap=avail_paths,
                                )
                            )
                            d_artifact.relative_path = content_artifact.relative_path
                        else:
                            msg = _(
                                "No declared artifact with relative path '{rp}' for content '{c}'"
                                " from remote '{rname}', and no paths available."
                            )
                            raise ValueError(
                                msg.format(
                                    rp=d_artifact.relative_path,
                                    c=d_content.content.natural_key(),
                                    rname=d_artifact.remote.name,
                                )
                            )
                    else:
                        msg = _('No declared artifact with relative path "{rp}" for content "{c}"')
                        raise ValueError(
                            msg.format(rp=d_artifact.relative_path, c=d_content.content)
                        )

                async for remote_artifact in sync_to_async_iterable(
                    content_artifact._remote_artifact_saver_ras
                ):
                    if d_artifact.url == remote_artifact.url:
                        break

                    if d_artifact.remote.pk == remote_artifact.remote_id:
                        key = f"{content_artifact.pk}-{remote_artifact.remote_id}"
                        remote_artifact.url = d_artifact.url
                        ras_to_update[key] = remote_artifact
                        break
                else:
                    remote_artifact = self._create_remote_artifact(d_artifact, content_artifact)
                    key = f"{content_artifact.pk}-{d_artifact.remote.pk}"
                    ras_to_create[key] = remote_artifact

        # Make sure we create/update RemoteArtifacts in a stable order, to help
        # prevent deadlocks in high-concurrency environments. We can rely on the
        # Artifact sha256 for our ordering.
        if ras_to_create:
            ras_to_create_ordered = sorted(list(ras_to_create.values()), key=lambda x: x.sha256)
            await sync_to_async(RemoteArtifact.objects.bulk_create)(ras_to_create_ordered)
        if ras_to_update:
            ras_to_update_ordered = sorted(list(ras_to_update.values()), key=lambda x: x.sha256)
            await sync_to_async(RemoteArtifact.objects.bulk_update)(
                ras_to_update_ordered, fields=["url"]
            )

    @staticmethod
    def _create_remote_artifact(d_artifact, content_artifact):
        ra = RemoteArtifact(
            url=d_artifact.url,
            size=d_artifact.artifact.size,
            md5=d_artifact.artifact.md5,
            sha1=d_artifact.artifact.sha1,
            sha224=d_artifact.artifact.sha224,
            sha256=d_artifact.artifact.sha256,
            sha384=d_artifact.artifact.sha384,
            sha512=d_artifact.artifact.sha512,
            content_artifact=content_artifact,
            remote=d_artifact.remote,
        )
        ra.validate_checksums()
        return ra


class ACSArtifactHandler(Stage):
    """
    API stage to download :class:`~pulpcore.plugin.models.Artifact` files from Alternate
    Content Source if available.
    """

    async def run(self):
        async for batch in self.batches():
            acs_exists = await sync_to_async(AlternateContentSource.objects.exists)()
            if acs_exists:
                # Gather batch d_artifact checksums
                batch_checksums = defaultdict(list)
                for d_content in batch:
                    for d_artifact in d_content.d_artifacts:
                        for cks_type in d_artifact.artifact.COMMON_DIGEST_FIELDS:
                            if getattr(d_artifact.artifact, cks_type):
                                batch_checksums[cks_type].append(
                                    getattr(d_artifact.artifact, cks_type)
                                )

                batch_query = Q()
                for checksum_type in batch_checksums.keys():
                    batch_query.add(
                        Q(**{f"{checksum_type}__in": batch_checksums[checksum_type]}), Q.OR
                    )

                existing_ras = await sync_to_async(list)(
                    (
                        RemoteArtifact.objects.acs()
                        .filter(batch_query)
                        .only("url", "remote")
                        .select_related("remote")
                    )
                )
                existing_ras_dict = dict()
                for ra in existing_ras:
                    for c_type in Artifact.COMMON_DIGEST_FIELDS:
                        checksum = await sync_to_async(getattr)(ra, c_type)
                        if checksum:
                            existing_ras_dict[checksum] = {
                                "remote": ra.remote,
                                "url": ra.url,
                            }

                for d_content in batch:
                    for d_artifact in d_content.d_artifacts:
                        for checksum_type in Artifact.COMMON_DIGEST_FIELDS:
                            if getattr(d_artifact.artifact, checksum_type):
                                checksum = getattr(d_artifact.artifact, checksum_type)
                                if checksum in existing_ras_dict:
                                    d_artifact.url = existing_ras_dict[checksum]["url"]
                                    d_artifact.remote = existing_ras_dict[checksum]["remote"]

            for d_content in batch:
                await self.put(d_content)



from unittest import TestCase
from types import SimpleNamespace

import mock
from rest_framework import serializers

from pulpcore.app.serializers import (
    PublicationSerializer,
    RemoteSerializer,
)


class TestRemoteSerializer(TestCase):
    minimal_data = {"name": "test", "url": "http://whatever"}

    def test_validate_proxy_creds_update(self):
        Remote = SimpleNamespace(
            proxy_url="http://whatever",
            proxy_username="user",
            proxy_password="pass",
            **self.minimal_data,
        )
        data = {"proxy_username": "user42"}
        serializer = RemoteSerializer(Remote, data=data, partial=True)
        serializer.is_valid(raise_exception=True)

    def test_validate_proxy_creds_update_invalid(self):
        Remote = SimpleNamespace(
            proxy_url="http://whatever",
            proxy_username=None,
            proxy_password=None,
            **self.minimal_data,
        )
        data = {"proxy_username": "user"}
        serializer = RemoteSerializer(Remote, data=data, partial=True)
        with self.assertRaises(serializers.ValidationError) as ctx:
            serializer.is_valid(raise_exception=True)
        self.assertIn("can only be specified together", str(ctx.exception))


class TestPublicationSerializer(TestCase):
    @mock.patch("pulpcore.app.serializers.repository.models.Repository")
    def test_validate_repository_only(self, mock_repo):
        data = {"repository": mock_repo}
        serializer = PublicationSerializer()
        new_data = serializer.validate(data)
        self.assertEqual(new_data, {"repository_version": mock_repo.latest_version.return_value})
        mock_repo.latest_version.assert_called_once_with()

    def test_validate_repository_version_only(self):
        mock_version = mock.MagicMock()
        data = {"repository_version": mock_version}
        serializer = PublicationSerializer()
        new_data = serializer.validate(data)
        self.assertEqual(new_data, {"repository_version": mock_version})

    def test_validate_repository_and_repository_version(self):
        mock_version = mock.MagicMock()
        mock_repository = mock.MagicMock()
        data = {"repository_version": mock_version, "repository": mock_repository}
        serializer = PublicationSerializer()
        with self.assertRaises(serializers.ValidationError):
            serializer.validate(data)

    def test_validate_no_repository_no_version(self):
        serializer = PublicationSerializer()
        with self.assertRaises(serializers.ValidationError):
            serializer.validate({})

    @mock.patch("pulpcore.app.serializers.repository.models.RepositoryVersion")
    def test_validate_repository_only_unknown_field(self, mock_version):
        mock_repo = mock.MagicMock()
        data = {"repository": mock_repo, "unknown_field": "unknown"}
        serializer = PublicationSerializer(data=data)
        with self.assertRaises(serializers.ValidationError):
            serializer.validate(data)

    def test_validate_repository_version_only_unknown_field(self):
        mock_version = mock.MagicMock()
        data = {"repository_version": mock_version, "unknown_field": "unknown"}
        serializer = PublicationSerializer(data=data)
        with self.assertRaises(serializers.ValidationError):
            serializer.validate(data)



# Generated by Django 3.2.9 on 2021-12-06 16:28

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0082_add_manage_roles_permissions'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='group',
            options={'permissions': [('manage_roles_group', 'Can manage role assignments on group')]},
        ),
    ]



# Generated by Django 2.2.6 on 2019-11-04 20:00

from django.db import migrations, models
import uuid


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0011_relative_path'),
    ]

    operations = [
        migrations.CreateModel(
            name='FileSystemExporter',
            fields=[
                ('pulp_id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('pulp_created', models.DateTimeField(auto_now_add=True)),
                ('pulp_last_updated', models.DateTimeField(auto_now=True, null=True)),
                ('pulp_type', models.TextField(default=None)),
                ('name', models.CharField(db_index=True, max_length=255, unique=True)),
                ('path', models.TextField()),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.DeleteModel(
            name='Exporter',
        ),
    ]



from importlib import import_module
from unittest.mock import Mock

from django.test import TestCase

from pulpcore.app.models import Remote


class TestMoveRemoteUrlCredentialsMigration(TestCase):
    """Test the move remote url credentials migration."""

    def setUp(self):
        self.remote = Remote.objects.create(
            name="test-url", url="http://elladan:lembas@rivendell.org", pulp_type="file"
        )

    def tearDown(self):
        self.remote.delete()

    def test_move_remote_url_credentials(self):
        apps_mock = Mock()
        apps_mock.get_model = Mock(return_value=Remote)

        # use import_module due to underscores
        migration = import_module("pulpcore.app.migrations.0077_move_remote_url_credentials")
        migration.move_remote_url_credentials(apps_mock, None)

        self.remote = Remote.objects.get(name="test-url")
        self.assertEqual(self.remote.url, "http://rivendell.org")
        self.assertEqual(self.remote.username, "elladan")
        self.assertEqual(self.remote.password, "lembas")



"""Tests related to content path."""
import unittest

from pulp_smash import api, config
from pulp_smash.pulp3.bindings import delete_orphans
from pulp_smash.pulp3.utils import gen_remote, gen_repo, sync

from pulpcore.tests.functional.api.using_plugin.constants import (
    FILE_FIXTURE_MANIFEST_URL,
    FILE_REMOTE_PATH,
    FILE_REPO_PATH,
)
from pulpcore.tests.functional.api.using_plugin.utils import create_file_publication
from pulpcore.tests.functional.api.using_plugin.utils import (  # noqa:F401
    set_up_module as setUpModule,
)


class SyncPublishContentPathTestCase(unittest.TestCase):
    """Test whether sync/publish for content already in Pulp.

    Different code paths are used in Pulp for the cases when artifacts are
    already present on the filesystem during sync and when they are not
    downloaded yet

    This test targets the following issue:

    `Pulp #4442 <https://pulp.plan.io/issues/4442>`_

    Does the following:

    1. Assure that no content from repository A is downloaded.
    2. Sync/publish repository A with download policy immediate.
    3. Sync/publish repository A again with download policy immediate.
    4. No failure in 2 shows that sync went fine when content was
       not present on the disk and in the database.
    5. No failure in 3 shows that sync went fine when content was already
       present on the disk and in the database.

    """

    def test_all(self):
        """Test whether sync/publish for content already in Pulp."""
        cfg = config.get_config()
        client = api.Client(cfg, api.page_handler)

        # step 1. delete orphans to assure that no content is present on disk,
        # or database.
        delete_orphans()

        remote = client.post(FILE_REMOTE_PATH, gen_remote(FILE_FIXTURE_MANIFEST_URL))
        self.addCleanup(client.delete, remote["pulp_href"])

        repo = client.post(FILE_REPO_PATH, gen_repo())
        self.addCleanup(client.delete, repo["pulp_href"])

        for _ in range(2):
            sync(cfg, remote, repo)
            repo = client.get(repo["pulp_href"])
            create_file_publication(cfg, repo)



# Generated by Django 2.2.18 on 2021-03-03 10:59

import contextlib

from urllib.parse import urlparse, urlunparse

from django.db import migrations


def move_proxy_creds_up(apps, schema_editor):
    Remote = apps.get_model('core', 'Remote')
    for remote in Remote.objects.all():
        if remote.proxy_url and "@" in remote.proxy_url:
            with contextlib.suppress(Exception):
                parsed_url = urlparse(remote.proxy_url)
                netloc = parsed_url.netloc
                auth, url = netloc.rsplit("@", maxsplit=1)
                remote.proxy_username, remote.proxy_password = auth.split(":", maxsplit=1)
                remote.proxy_url = urlunparse(parsed_url._replace(netloc=url))
                remote.save()


def move_proxy_creds_down(apps, schema_editor):
    Remote = apps.get_model('core', 'Remote')
    for remote in Remote.objects.all():
        if remote.proxy_url and remote.proxy_username and remote.proxy_password:
            with contextlib.suppress(Exception):
                parsed_url = urlparse(remote.proxy_url)
                url = parsed_url.netloc
                netloc = remote.proxy_username + ":" + remote.proxy_password + "@" + url
                remote.proxy_username = None
                remote.proxy_password = None
                remote.proxy_url = urlunparse(parsed_url._replace(netloc=netloc))
                remote.save()


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0059_proxy_creds'),
    ]

    operations = [
        migrations.RunPython(move_proxy_creds_up, reverse_code=move_proxy_creds_down, elidable=True),
    ]



# Generated by Django 2.2.5 on 2019-09-18 15:52

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0007_delete_progress_proxies'),
    ]

    operations = [
        migrations.AlterUniqueTogether(
            name='publishedmetadata',
            unique_together={('publication', 'relative_path')},
        ),
        migrations.RemoveField(
            model_name='publishedmetadata',
            name='_created',
        ),
        migrations.RemoveField(
            model_name='publishedmetadata',
            name='_id',
        ),
        migrations.RemoveField(
            model_name='publishedmetadata',
            name='_last_updated',
        ),
        migrations.RemoveField(
            model_name='publishedmetadata',
            name='file',
        ),
        migrations.AddField(
            model_name='publishedmetadata',
            name='content_ptr',
            field=models.OneToOneField(auto_created=True, default=0,
                                       on_delete=django.db.models.deletion.CASCADE,
                                       parent_link=True, primary_key=True,
                                       related_name='published_metadata',
                                       serialize=False, to='core.Content'),
            preserve_default=False,
        ),
    ]



from asyncio import Queue
import pathlib
import time
import uuid

from pulpcore.app.models import Task


CONN = None


class ProfilingQueue(Queue):
    """
    A customized subclass of asyncio.Queue that records time in the queue and between queues.

    This Profiler records some data on items that are inserted and removed from Queues. This data is
    stored on items in a dictionary attribute called 'extra_data'. If this attribute does not exist
    on an item, the ProfileQueue adds it.

    The following statistics are computed for each Queue and the stage that it feeds into:

        * waiting time - The number of seconds an item waited in the Queue for this stage.
        * service time - The number of seconds an item received service in this stage.
        * queue_length - The number of waiting items in the queue, measured before each new arrival.
        * interarrival_time - The number of seconds since the previous arrival to this Queue.

    See the :meth:`create_profile_db_and_connection()` docs for more info on the database tables and
    layout.

    Args:
         stage_uuid (uuid.UUID): The uuid of the stage this ProfilingQueue delivers work into.
         args (tuple): unused positional arguments
         kwargs (dict): unused keyword arguments
    """

    def __init__(self, stage_uuid, *args, **kwargs):
        self.last_arrival_time = time.time()
        self.stage_uuid = stage_uuid
        return super().__init__(*args, **kwargs)

    def get_nowait(self):
        """
        Thinly wrap `asyncio.get_nowait` and record when get_nowait() operations happen.
        """
        item = super().get_nowait()
        if item:
            now = time.time()
            item.extra_data["last_waiting_time"] = now - item.extra_data["lastput_time"]
            item.extra_data["last_get_time"] = now
        return item

    def put_nowait(self, item):
        """
        Thinly wrap `asyncio.put_nowait` happens and write statistics about them to the sqlite3 DB.

        This method computes and writes the following statistics: waiting time, service time, queue
        length, and interarrival time.
        """
        if item:
            now = time.time()
            if not hasattr(item, "extra_data"):
                # track stages that use QuerySet items too
                item.extra_data = {}
            try:
                last_waiting_time = item.extra_data["last_waiting_time"]
            except KeyError:
                pass
            else:
                service_time = now - item.extra_data["last_get_time"]
                sql = (
                    "INSERT INTO traffic (uuid, waiting_time, service_time) VALUES ("
                    "'{uuid}','{waiting_time}','{service_time}')"
                )
                formatted_sql = sql.format(
                    uuid=self.stage_uuid, waiting_time=last_waiting_time, service_time=service_time
                )
                CONN.cursor().execute(formatted_sql)

            interarrival_time = now - self.last_arrival_time
            sql = (
                "INSERT INTO system (uuid, length, interarrival_time) VALUES ("
                "'{uuid}','{length}','{interarrival}')"
            )
            formatted_sql = sql.format(
                uuid=self.stage_uuid, length=super().qsize(), interarrival=interarrival_time
            )
            CONN.cursor().execute(formatted_sql)
            CONN.commit()

            item.extra_data["lastput_time"] = now
            self.last_arrival_time = now
        return super().put_nowait(item)

    @staticmethod
    def make_and_record_queue(stage, num, maxsize):
        """
        Create a ProfileQueue that is associated with the stage it feeds and record it in sqlite3.

        Args:
            stage (uuid.UUID): The uuid of this stage for correlation with other table data.
            num: (int): The number in the pipeline this stage is at, starting from 0, 1, etc.
            maxsize: The `maxsize` parameter being used to configure the ProfilingQueue with.

        Returns:
            ProfilingQueue: The configured ProfilingQueue that was also recorded in the db.
        """
        if CONN is None:
            create_profile_db_and_connection()
        stage_id = uuid.uuid4()
        stage_name = ".".join([stage.__class__.__module__, stage.__class__.__name__])
        sql = "INSERT INTO stages (uuid, name, num) VALUES ('{uuid}','{stage}','{num}')"
        formatted_sql = sql.format(uuid=stage_id, stage=stage_name, num=num)
        CONN.cursor().execute(formatted_sql)
        in_q = ProfilingQueue(stage_id, maxsize=maxsize)
        CONN.commit()
        return in_q


def create_profile_db_and_connection():
    """
    Create a profile db from this tasks UUID and a sqlite3 connection to that databases.

    The database produced has three tables with the following SQL format:

    The `stages` table stores info about the pipeline itself and stores 3 fields
    * uuid - the uuid of the stage
    * name - the name of the stage
    * num - the number of the stage starting at 0

    The `traffic` table stores 3 fields:
    * uuid - the uuid of the stage this queue feeds into
    * waiting_time - the amount of time the item is waiting in the queue before it enters the stage.
    * service_time - the service time the item spent in the stage.

    The `system` table stores 3 fields:
    * uuid - The uuid of stage this queue feeds into
    * length - The length of items in this queue, measured just before each arrival.
    * interarrival_time - The amount of time since the last arrival.
    """
    debug_data_dir = "/var/lib/pulp/debug/"
    pathlib.Path(debug_data_dir).mkdir(parents=True, exist_ok=True)
    current_task = Task.current()
    if current_task:
        db_path = debug_data_dir + str(current_task.pk)
    else:
        db_path = debug_data_dir + str(uuid.uuid4())

    import sqlite3

    global CONN
    CONN = sqlite3.connect(db_path)
    c = CONN.cursor()

    # Create table
    c.execute(
        """CREATE TABLE stages
                 (uuid varchar(36), name text, num int)"""
    )

    # Create table
    c.execute(
        """CREATE TABLE traffic
                 (uuid varchar(36), waiting_time real, service_time real)"""
    )

    # Create table
    c.execute(
        """CREATE TABLE system
                 (uuid varchar(36), length int, interarrival_time real)"""
    )

    return CONN



import logging
import os
from gettext import gettext as _

from cryptography.fernet import Fernet
from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.db.models import FileField, Lookup
from django.db.models.fields import Field, TextField
from django.utils.encoding import force_bytes, force_str
from django.utils.functional import cached_property


from pulpcore.app.files import TemporaryDownloadedFile

_logger = logging.getLogger(__name__)


class ArtifactFileField(FileField):
    """
    A custom FileField that always saves files to location specified by 'upload_to'.

    The field can be set as either a path to the file or File object. In both cases the file is
    moved or copied to the location specified by 'upload_to' field parameter.
    """

    def pre_save(self, model_instance, add):
        """
        Return FieldFile object which specifies path to the file to be stored in database.

        There are two ways to get artifact into Pulp: sync and upload.

        The upload case
         - file is not stored yet, aka file._committed = False
         - nothing to do here in addition to Django pre_save actions

        The sync case:
         - file is already stored in a temporary location, aka file._committed = True
         - it needs to be moved into Pulp artifact storage if it's not there
         - TemporaryDownloadedFile takes care of correctly set storage path
         - only then Django pre_save actions should be performed

        Args:
            model_instance (`class::pulpcore.plugin.Artifact`): The instance this field belongs to.
            add (bool): Whether the instance is being saved to the database for the first time.
                        Ignored by Django pre_save method.

        Returns:
            FieldFile object just before saving.

        """
        file = model_instance.file
        artifact_storage_path = self.upload_to(model_instance, "")

        already_in_place = file.name in [
            artifact_storage_path,
            os.path.join(settings.MEDIA_ROOT, artifact_storage_path),
        ]
        is_in_artifact_storage = file.name.startswith(os.path.join(settings.MEDIA_ROOT, "artifact"))

        if not already_in_place and is_in_artifact_storage:
            raise ValueError(
                _(
                    "The file referenced by the Artifact is already present in "
                    "Artifact storage. Files must be stored outside this location "
                    "prior to Artifact creation."
                )
            )

        move = file._committed and file.name != artifact_storage_path
        if move:
            if not already_in_place:
                file._file = TemporaryDownloadedFile(open(file.name, "rb"))
            file._committed = False

        return super().pre_save(model_instance, add)


class EncryptedTextField(TextField):
    """A field mixin that encrypts text using settings.DB_ENCRYPTION_KEY."""

    def __init__(self, *args, **kwargs):
        if kwargs.get("primary_key"):
            raise ImproperlyConfigured("EncryptedTextField does not support primary_key=True.")
        if kwargs.get("unique"):
            raise ImproperlyConfigured("EncryptedTextField does not support unique=True.")
        if kwargs.get("db_index"):
            raise ImproperlyConfigured("EncryptedTextField does not support db_index=True.")
        super().__init__(*args, **kwargs)

    @cached_property
    def _fernet(self):
        _logger.debug(f"Loading encryption key from {settings.DB_ENCRYPTION_KEY}")
        with open(settings.DB_ENCRYPTION_KEY, "rb") as key_file:
            return Fernet(key_file.read())

    def get_db_prep_save(self, value, connection):
        value = super().get_db_prep_save(value, connection)
        if value is not None:
            return force_str(self._fernet.encrypt(force_bytes(value)))

    def from_db_value(self, value, expression, connection):
        if value is not None:
            return force_str(self._fernet.decrypt(force_bytes(value)))


@Field.register_lookup
class NotEqualLookup(Lookup):
    # this is copied from https://docs.djangoproject.com/en/3.0/howto/custom-lookups/
    lookup_name = "ne"

    def as_sql(self, compiler, connection):
        lhs, lhs_params = self.process_lhs(compiler, connection)
        rhs, rhs_params = self.process_rhs(compiler, connection)
        params = lhs_params + rhs_params
        return "%s <> %s" % (lhs, rhs), params



"""Test that operations can be performed over tasks."""
import unittest

from pulp_smash import api, config, utils
from pulp_smash.pulp3.constants import (
    BASE_DISTRIBUTION_PATH,
    P3_TASK_END_STATES,
    TASKS_PATH,
)
from pulp_smash.pulp3.utils import gen_repo, gen_distribution, get_content, modify_repo, sync
from requests import HTTPError

from pulpcore.tests.functional.api.using_plugin.constants import (
    FILE_CONTENT_NAME,
    FILE_REMOTE_PATH,
    FILE_REPO_PATH,
)
from pulpcore.tests.functional.utils import set_up_module as setUpModule  # noqa:F401
from pulpcore.tests.functional.utils import skip_if
from pulpcore.tests.functional.api.using_plugin.utils import gen_file_remote

_DYNAMIC_TASKS_ATTRS = ("finished_at",)
"""Task attributes that are dynamically set by Pulp, not set by a user."""


class TasksTestCase(unittest.TestCase):
    """Perform different operation over tasks.

    This test targets the following issues:

    * `Pulp #3144 <https://pulp.plan.io/issues/3144>`_
    * `Pulp #3527 <https://pulp.plan.io/issues/3527>`_
    * `Pulp Smash #754 <https://github.com/pulp/pulp-smash/issues/754>`_
    """

    @classmethod
    def setUpClass(cls):
        """Create class-wide variables."""
        cls.client = api.Client(config.get_config(), api.json_handler)
        cls.task = {}

    def test_01_create_task(self):
        """Create a task."""
        repo = self.client.post(FILE_REPO_PATH, gen_repo())
        self.addCleanup(self.client.delete, repo["pulp_href"])
        attrs = {"description": utils.uuid4()}
        response = self.client.patch(repo["pulp_href"], attrs)
        self.task.update(self.client.get(response["task"]))

    @skip_if(bool, "task", False)
    def test_02_read_href(self):
        """Read a task by its pulp_href."""
        task = self.client.get(self.task["pulp_href"])
        for key, val in self.task.items():
            if key in _DYNAMIC_TASKS_ATTRS:
                continue
            with self.subTest(key=key):
                self.assertEqual(task[key], val, task)

    @skip_if(bool, "task", False)
    def test_02_read_href_with_specific_fields(self):
        """Read a task by its pulp_hrefproviding specific fields."""
        fields = ("pulp_href", "state", "worker")
        task = self.client.get(self.task["pulp_href"], params={"fields": ",".join(fields)})
        self.assertEqual(sorted(fields), sorted(task.keys()))

    @skip_if(bool, "task", False)
    def test_02_read_task_without_specific_fields(self):
        """Read a task by its href excluding specific fields."""
        # requests doesn't allow the use of != in parameters.
        url = "{}?exclude_fields=state".format(self.task["pulp_href"])
        task = self.client.get(url)
        self.assertNotIn("state", task.keys())

    @skip_if(bool, "task", False)
    def test_02_read_task_with_minimal_fields(self):
        """Read a task by its href filtering minimal fields."""
        task = self.client.get(self.task["pulp_href"], params={"minimal": True})
        response_fields = task.keys()
        self.assertNotIn("progress_reports", response_fields)
        self.assertNotIn("parent_task", response_fields)
        self.assertNotIn("error", response_fields)

    @skip_if(bool, "task", False)
    def test_02_read_invalid_worker(self):
        """Read a task using an invalid worker name."""
        with self.assertRaises(HTTPError):
            self.filter_tasks({"worker": utils.uuid4()})

    @skip_if(bool, "task", False)
    def test_02_read_valid_worker(self):
        """Read a task using a valid worker name."""
        page = self.filter_tasks({"worker": self.task["worker"]})
        self.assertNotEqual(len(page["results"]), 0, page["results"])

    def test_02_read_invalid_date(self):
        """Read a task by an invalid date."""
        with self.assertRaises(HTTPError):
            self.filter_tasks({"finished_at": utils.uuid4(), "started_at": utils.uuid4()})

    @skip_if(bool, "task", False)
    def test_02_read_valid_date(self):
        """Read a task by a valid date."""
        page = self.filter_tasks({"started_at": self.task["started_at"]})
        self.assertGreaterEqual(len(page["results"]), 1, page["results"])

    @skip_if(bool, "task", False)
    def test_02_search_task_by_name(self):
        """Search Task by its name.

        This test targets the following issue:

        * `Pulp #4230 <https://pulp.plan.io/issues/4230>`_

        Do the following:

        1. Assert that task has a field name, and that this field is not empty.
        2. Filter the tasks by name.
        3. Assert the created task is included on the search results.
        """
        # step 1
        self.assertIsNotNone(self.task.get("name"))
        # step 2
        search_results = self.filter_tasks({"name": self.task["name"]})
        # step 3
        self.assertIn(self.task, search_results["results"])

    def test_02_search_by_invalid_name(self):
        """Search passing invalid name and assert nothing is returned."""
        search_results = self.filter_tasks({"name": "this-is-not-a-task-name"})
        self.assertEqual(search_results["count"], 0)
        self.assertEqual(len(search_results["results"]), 0)

    @skip_if(bool, "task", False)
    def test_03_delete_tasks(self):
        """Delete a task."""
        # If this assertion fails, then either Pulp's tasking system or Pulp
        # Smash's code for interacting with the tasking system has a bug.
        self.assertIn(self.task["state"], P3_TASK_END_STATES)
        self.client.delete(self.task["pulp_href"])
        with self.assertRaises(HTTPError):
            self.client.get(self.task["pulp_href"])

    def filter_tasks(self, criteria):
        """Filter tasks based on the provided criteria."""
        return self.client.get(TASKS_PATH, params=criteria)


class FilterTaskCreatedResourcesTestCase(unittest.TestCase):
    """Perform filtering over the task's field created_resources.

    This test targets the following issue:

    * `Pulp #5180 <https://pulp.plan.io/issues/5180>`_
    """

    def test_read_fields_created_resources_only(self):
        """Read created resources from the requested fields."""
        client = api.Client(config.get_config(), api.page_handler)
        distribution_path = "{}file/file/".format(BASE_DISTRIBUTION_PATH)
        response = client.post(distribution_path, gen_distribution())

        task = client.get(response["task"])
        self.addCleanup(client.delete, task["created_resources"][0])

        filtered_task = client.get(task["pulp_href"], params={"fields": "created_resources"})

        self.assertEqual(len(filtered_task), 1, filtered_task)

        self.assertEqual(
            task["created_resources"], filtered_task["created_resources"], filtered_task
        )


class FilterTaskResourcesTestCase(unittest.TestCase):
    """Perform filtering over reserved resources.

    This test targets the following issue:
    * `Pulp #5120 <https://pulp.plan.io/issues/5120>`_

    Perform filtering for contents of created resources.

    This test targets the following issue:
    * `Pulp #4931 <https://pulp.plan.io/issues/4931>`_
    """

    @classmethod
    def setUpClass(cls):
        """Create class-wide variables."""
        cls.cfg = config.get_config()
        cls.client = api.Client(cls.cfg, api.page_handler)
        cls.remote = cls.client.post(FILE_REMOTE_PATH, gen_file_remote())
        cls.repository = cls.client.post(FILE_REPO_PATH, gen_repo())
        response = sync(cls.cfg, cls.remote, cls.repository)
        cls.created_repo_version = response["pulp_href"]
        cls.repository = cls.client.get(cls.repository["pulp_href"])
        for file_content in get_content(cls.repository)[FILE_CONTENT_NAME]:
            modify_repo(cls.cfg, cls.repository, remove_units=[file_content])
        attrs = {"description": utils.uuid4()}
        response = cls.client.patch(cls.repository["pulp_href"], attrs)
        cls.repo_update_task = cls.client.get(response["task"])

    @classmethod
    def tearDownClass(cls):
        """Clean created resources."""
        cls.client.delete(cls.repository["pulp_href"])
        cls.client.delete(cls.remote["pulp_href"])
        cls.client.delete(cls.repo_update_task["pulp_href"])

    def test_01_filter_tasks_by_reserved_resources(self):
        """Filter all tasks by a particular reserved resource."""
        filter_params = {
            "reserved_resources_record": self.repo_update_task["reserved_resources_record"][0]
        }
        results = self.client.get(TASKS_PATH, params=filter_params)
        self.assertEqual(len(results), 5, results)
        self.assertEqual(self.repo_update_task, results[0], results)

    def test_02_filter_tasks_by_non_existing_resources(self):
        """Filter all tasks by a non-existing reserved resource."""
        filter_params = {"reserved_resources_record": "a_resource_should_be_never_named_like_this"}
        with self.assertRaises(HTTPError):
            self.client.get(TASKS_PATH, params=filter_params)

    def test_03_filter_tasks_by_created_resources(self):
        """Filter all tasks by a particular created resource."""
        filter_params = {"created_resources": self.created_repo_version}
        results = self.client.get(TASKS_PATH, params=filter_params)
        self.assertEqual(len(results), 1, results)
        self.assertEqual([self.created_repo_version], results[0]["created_resources"], results)

    def test_04_filter_tasks_by_non_existing_resources(self):
        """Filter all tasks by a non-existing reserved resource."""
        filter_params = {"created_resources": "a_resource_should_be_never_named_like_this"}
        with self.assertRaises(HTTPError):
            self.client.get(TASKS_PATH, params=filter_params)



# Generated by Django 2.2.2 on 2019-07-05 17:06

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='artifact',
            name='size',
            field=models.BigIntegerField(),
        ),
    ]



# Generated by Django 2.2.6 on 2019-10-01 19:31

from django.db import migrations, models
import django.db.models.deletion
import pulpcore.app.models.task


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0009_remove_task_non_fatal_errors'),
    ]

    operations = [
        migrations.RenameField(
            model_name='artifact',
            old_name='_created',
            new_name='pulp_created',
        ),
        migrations.RenameField(
            model_name='artifact',
            old_name='_id',
            new_name='pulp_id',
        ),
        migrations.RenameField(
            model_name='artifact',
            old_name='_last_updated',
            new_name='pulp_last_updated',
        ),
        migrations.RenameField(
            model_name='basedistribution',
            old_name='_created',
            new_name='pulp_created',
        ),
        migrations.RenameField(
            model_name='basedistribution',
            old_name='_id',
            new_name='pulp_id',
        ),
        migrations.RenameField(
            model_name='basedistribution',
            old_name='_last_updated',
            new_name='pulp_last_updated',
        ),
        migrations.RenameField(
            model_name='basedistribution',
            old_name='_type',
            new_name='pulp_type',
        ),
        migrations.RenameField(
            model_name='content',
            old_name='_created',
            new_name='pulp_created',
        ),
        migrations.RenameField(
            model_name='content',
            old_name='_id',
            new_name='pulp_id',
        ),
        migrations.RenameField(
            model_name='content',
            old_name='_last_updated',
            new_name='pulp_last_updated',
        ),
        migrations.RenameField(
            model_name='content',
            old_name='_type',
            new_name='pulp_type',
        ),
        migrations.RenameField(
            model_name='contentappstatus',
            old_name='_created',
            new_name='pulp_created',
        ),
        migrations.RenameField(
            model_name='contentappstatus',
            old_name='_id',
            new_name='pulp_id',
        ),
        migrations.RenameField(
            model_name='contentappstatus',
            old_name='_last_updated',
            new_name='pulp_last_updated',
        ),
        migrations.RenameField(
            model_name='contentartifact',
            old_name='_created',
            new_name='pulp_created',
        ),
        migrations.RenameField(
            model_name='contentartifact',
            old_name='_id',
            new_name='pulp_id',
        ),
        migrations.RenameField(
            model_name='contentartifact',
            old_name='_last_updated',
            new_name='pulp_last_updated',
        ),
        migrations.RenameField(
            model_name='contentguard',
            old_name='_created',
            new_name='pulp_created',
        ),
        migrations.RenameField(
            model_name='contentguard',
            old_name='_id',
            new_name='pulp_id',
        ),
        migrations.RenameField(
            model_name='contentguard',
            old_name='_last_updated',
            new_name='pulp_last_updated',
        ),
        migrations.RenameField(
            model_name='contentguard',
            old_name='_type',
            new_name='pulp_type',
        ),
        migrations.RenameField(
            model_name='createdresource',
            old_name='_created',
            new_name='pulp_created',
        ),
        migrations.RenameField(
            model_name='createdresource',
            old_name='_id',
            new_name='pulp_id',
        ),
        migrations.RenameField(
            model_name='createdresource',
            old_name='_last_updated',
            new_name='pulp_last_updated',
        ),
        migrations.RenameField(
            model_name='exporter',
            old_name='_created',
            new_name='pulp_created',
        ),
        migrations.RenameField(
            model_name='exporter',
            old_name='_id',
            new_name='pulp_id',
        ),
        migrations.RenameField(
            model_name='exporter',
            old_name='_last_updated',
            new_name='pulp_last_updated',
        ),
        migrations.RenameField(
            model_name='exporter',
            old_name='_type',
            new_name='pulp_type',
        ),
        migrations.RenameField(
            model_name='progressreport',
            old_name='_created',
            new_name='pulp_created',
        ),
        migrations.RenameField(
            model_name='progressreport',
            old_name='_id',
            new_name='pulp_id',
        ),
        migrations.RenameField(
            model_name='progressreport',
            old_name='_last_updated',
            new_name='pulp_last_updated',
        ),
        migrations.RenameField(
            model_name='publication',
            old_name='_created',
            new_name='pulp_created',
        ),
        migrations.RenameField(
            model_name='publication',
            old_name='_id',
            new_name='pulp_id',
        ),
        migrations.RenameField(
            model_name='publication',
            old_name='_last_updated',
            new_name='pulp_last_updated',
        ),
        migrations.RenameField(
            model_name='publication',
            old_name='_type',
            new_name='pulp_type',
        ),
        migrations.RenameField(
            model_name='publishedartifact',
            old_name='_created',
            new_name='pulp_created',
        ),
        migrations.RenameField(
            model_name='publishedartifact',
            old_name='_id',
            new_name='pulp_id',
        ),
        migrations.RenameField(
            model_name='publishedartifact',
            old_name='_last_updated',
            new_name='pulp_last_updated',
        ),
        migrations.RenameField(
            model_name='publisher',
            old_name='_created',
            new_name='pulp_created',
        ),
        migrations.RenameField(
            model_name='publisher',
            old_name='_id',
            new_name='pulp_id',
        ),
        migrations.RenameField(
            model_name='publisher',
            old_name='_last_updated',
            new_name='pulp_last_updated',
        ),
        migrations.RenameField(
            model_name='publisher',
            old_name='_type',
            new_name='pulp_type',
        ),
        migrations.RenameField(
            model_name='remote',
            old_name='_created',
            new_name='pulp_created',
        ),
        migrations.RenameField(
            model_name='remote',
            old_name='_id',
            new_name='pulp_id',
        ),
        migrations.RenameField(
            model_name='remote',
            old_name='_last_updated',
            new_name='pulp_last_updated',
        ),
        migrations.RenameField(
            model_name='remote',
            old_name='_type',
            new_name='pulp_type',
        ),
        migrations.RenameField(
            model_name='remoteartifact',
            old_name='_created',
            new_name='pulp_created',
        ),
        migrations.RenameField(
            model_name='remoteartifact',
            old_name='_id',
            new_name='pulp_id',
        ),
        migrations.RenameField(
            model_name='remoteartifact',
            old_name='_last_updated',
            new_name='pulp_last_updated',
        ),
        migrations.RenameField(
            model_name='repository',
            old_name='_created',
            new_name='pulp_created',
        ),
        migrations.RenameField(
            model_name='repository',
            old_name='_id',
            new_name='pulp_id',
        ),
        migrations.RenameField(
            model_name='repository',
            old_name='_last_updated',
            new_name='pulp_last_updated',
        ),
        migrations.RenameField(
            model_name='repositorycontent',
            old_name='_created',
            new_name='pulp_created',
        ),
        migrations.RenameField(
            model_name='repositorycontent',
            old_name='_id',
            new_name='pulp_id',
        ),
        migrations.RenameField(
            model_name='repositorycontent',
            old_name='_last_updated',
            new_name='pulp_last_updated',
        ),
        migrations.RenameField(
            model_name='repositoryversion',
            old_name='_created',
            new_name='pulp_created',
        ),
        migrations.RenameField(
            model_name='repositoryversion',
            old_name='_id',
            new_name='pulp_id',
        ),
        migrations.RenameField(
            model_name='repositoryversion',
            old_name='_last_updated',
            new_name='pulp_last_updated',
        ),
        migrations.RenameField(
            model_name='reservedresource',
            old_name='_created',
            new_name='pulp_created',
        ),
        migrations.RenameField(
            model_name='reservedresource',
            old_name='_id',
            new_name='pulp_id',
        ),
        migrations.RenameField(
            model_name='reservedresource',
            old_name='_last_updated',
            new_name='pulp_last_updated',
        ),
        migrations.RenameField(
            model_name='reservedresourcerecord',
            old_name='_created',
            new_name='pulp_created',
        ),
        migrations.RenameField(
            model_name='reservedresourcerecord',
            old_name='_id',
            new_name='pulp_id',
        ),
        migrations.RenameField(
            model_name='reservedresourcerecord',
            old_name='_last_updated',
            new_name='pulp_last_updated',
        ),
        migrations.RenameField(
            model_name='task',
            old_name='_created',
            new_name='pulp_created',
        ),
        migrations.RenameField(
            model_name='task',
            old_name='_id',
            new_name='pulp_id',
        ),
        migrations.RenameField(
            model_name='task',
            old_name='_last_updated',
            new_name='pulp_last_updated',
        ),
        migrations.RenameField(
            model_name='taskreservedresource',
            old_name='_created',
            new_name='pulp_created',
        ),
        migrations.RenameField(
            model_name='taskreservedresource',
            old_name='_id',
            new_name='pulp_id',
        ),
        migrations.RenameField(
            model_name='taskreservedresource',
            old_name='_last_updated',
            new_name='pulp_last_updated',
        ),
        migrations.RenameField(
            model_name='taskreservedresourcerecord',
            old_name='_created',
            new_name='pulp_created',
        ),
        migrations.RenameField(
            model_name='taskreservedresourcerecord',
            old_name='_id',
            new_name='pulp_id',
        ),
        migrations.RenameField(
            model_name='taskreservedresourcerecord',
            old_name='_last_updated',
            new_name='pulp_last_updated',
        ),
        migrations.RenameField(
            model_name='upload',
            old_name='_created',
            new_name='pulp_created',
        ),
        migrations.RenameField(
            model_name='upload',
            old_name='_id',
            new_name='pulp_id',
        ),
        migrations.RenameField(
            model_name='upload',
            old_name='_last_updated',
            new_name='pulp_last_updated',
        ),
        migrations.RenameField(
            model_name='uploadchunk',
            old_name='_created',
            new_name='pulp_created',
        ),
        migrations.RenameField(
            model_name='uploadchunk',
            old_name='_id',
            new_name='pulp_id',
        ),
        migrations.RenameField(
            model_name='uploadchunk',
            old_name='_last_updated',
            new_name='pulp_last_updated',
        ),
        migrations.RenameField(
            model_name='worker',
            old_name='_created',
            new_name='pulp_created',
        ),
        migrations.RenameField(
            model_name='worker',
            old_name='_id',
            new_name='pulp_id',
        ),
        migrations.RenameField(
            model_name='worker',
            old_name='_last_updated',
            new_name='pulp_last_updated',
        ),
        migrations.AlterField(
            model_name='basedistribution',
            name='content_guard',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, to='core.ContentGuard', to_field='pulp_id'),
        ),
        migrations.AlterField(
            model_name='basedistribution',
            name='remote',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, to='core.Remote', to_field='pulp_id'),
        ),
        migrations.AlterField(
            model_name='contentartifact',
            name='artifact',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.PROTECT, to='core.Artifact', to_field='pulp_id'),
        ),
        migrations.AlterField(
            model_name='contentartifact',
            name='content',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='core.Content', to_field='pulp_id'),
        ),
        migrations.AlterField(
            model_name='createdresource',
            name='task',
            field=models.ForeignKey(default=pulpcore.app.models.task.Task.current, on_delete=django.db.models.deletion.CASCADE, related_name='created_resources', to='core.Task', to_field='pulp_id'),
        ),
        migrations.AlterField(
            model_name='progressreport',
            name='task',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='progress_reports', to='core.Task', to_field='pulp_id'),
        ),
        migrations.AlterField(
            model_name='publication',
            name='publisher',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, to='core.Publisher', to_field='pulp_id'),
        ),
        migrations.AlterField(
            model_name='publication',
            name='repository_version',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='core.RepositoryVersion', to_field='pulp_id'),
        ),
        migrations.AlterField(
            model_name='publishedartifact',
            name='content_artifact',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='published_artifact', to='core.ContentArtifact', to_field='pulp_id'),
        ),
        migrations.AlterField(
            model_name='publishedartifact',
            name='publication',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='published_artifact', to='core.Publication', to_field='pulp_id'),
        ),
        migrations.AlterField(
            model_name='publishedmetadata',
            name='content_ptr',
            field=models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, related_name='published_metadata', serialize=False, to='core.Content', to_field='pulp_id'),
        ),
        migrations.AlterField(
            model_name='publishedmetadata',
            name='publication',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='published_metadata', to='core.Publication', to_field='pulp_id'),
        ),
        migrations.AlterField(
            model_name='remoteartifact',
            name='content_artifact',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='core.ContentArtifact', to_field='pulp_id'),
        ),
        migrations.AlterField(
            model_name='remoteartifact',
            name='remote',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='core.Remote', to_field='pulp_id'),
        ),
        migrations.AlterField(
            model_name='repositorycontent',
            name='content',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='version_memberships', to='core.Content', to_field='pulp_id'),
        ),
        migrations.AlterField(
            model_name='repositorycontent',
            name='repository',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='core.Repository', to_field='pulp_id'),
        ),
        migrations.AlterField(
            model_name='repositorycontent',
            name='version_added',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='added_memberships', to='core.RepositoryVersion', to_field='pulp_id'),
        ),
        migrations.AlterField(
            model_name='repositorycontent',
            name='version_removed',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, related_name='removed_memberships', to='core.RepositoryVersion', to_field='pulp_id'),
        ),
        migrations.AlterField(
            model_name='repositoryversion',
            name='base_version',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='versions', to='core.RepositoryVersion', to_field='pulp_id'),
        ),
        migrations.AlterField(
            model_name='repositoryversion',
            name='repository',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='versions', to='core.Repository', to_field='pulp_id'),
        ),
        migrations.AlterField(
            model_name='repositoryversioncontentdetails',
            name='repository_version',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='counts', to='core.RepositoryVersion', to_field='pulp_id'),
        ),
        migrations.AlterField(
            model_name='reservedresource',
            name='worker',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='reservations', to='core.Worker', to_field='pulp_id'),
        ),
        migrations.AlterField(
            model_name='task',
            name='parent',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='spawned_tasks', to='core.Task', to_field='pulp_id'),
        ),
        migrations.AlterField(
            model_name='task',
            name='worker',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='tasks', to='core.Worker', to_field='pulp_id'),
        ),
        migrations.AlterField(
            model_name='taskreservedresource',
            name='resource',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='core.ReservedResource', to_field='pulp_id'),
        ),
        migrations.AlterField(
            model_name='taskreservedresource',
            name='task',
            field=models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to='core.Task', to_field='pulp_id'),
        ),
        migrations.AlterField(
            model_name='taskreservedresourcerecord',
            name='resource',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='core.ReservedResourceRecord', to_field='pulp_id'),
        ),
        migrations.AlterField(
            model_name='taskreservedresourcerecord',
            name='task',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='core.Task', to_field='pulp_id'),
        ),
        migrations.AlterField(
            model_name='uploadchunk',
            name='upload',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='chunks', to='core.Upload', to_field='pulp_id'),
        ),
    ]



from gettext import gettext as _

from rest_framework import serializers

from pulpcore.app import models
from pulpcore.app.serializers import ModelSerializer


class ProgressReportSerializer(ModelSerializer):

    message = serializers.CharField(
        help_text=_("The message shown to the user for the progress report."), read_only=True
    )
    code = serializers.CharField(
        help_text=_("Identifies the type of progress report'."), read_only=True
    )
    state = serializers.CharField(
        help_text=_(
            "The current state of the progress report. The possible values are:"
            " 'waiting', 'skipped', 'running', 'completed', 'failed', 'canceled' and 'canceling'."
            " The default is 'waiting'."
        ),
        read_only=True,
    )
    total = serializers.IntegerField(help_text=_("The total count of items."), read_only=True)
    done = serializers.IntegerField(
        help_text=_("The count of items already processed. Defaults to 0."), read_only=True
    )
    suffix = serializers.CharField(
        help_text=_("The suffix to be shown with the progress report."),
        read_only=True,
        allow_null=True,
    )

    class Meta:
        model = models.ProgressReport
        # this serializer is meant to be nested inside Task serializer,
        # so it will not have its own endpoint, that's why
        # we need to explicitly define fields to exclude 'pulp_href' field.
        fields = ("message", "code", "state", "total", "done", "suffix")


class GroupProgressReportSerializer(ModelSerializer):

    message = serializers.CharField(
        help_text=_("The message shown to the user for the group progress report."), read_only=True
    )
    code = serializers.CharField(
        help_text=_("Identifies the type of group progress report'."), read_only=True
    )
    total = serializers.IntegerField(help_text=_("The total count of items."), read_only=True)
    done = serializers.IntegerField(
        help_text=_("The count of items already processed. Defaults to 0."), read_only=True
    )
    suffix = serializers.CharField(
        help_text=_("The suffix to be shown with the group progress report."),
        read_only=True,
        allow_null=True,
    )

    class Meta:
        model = models.GroupProgressReport
        # this serializer is meant to be nested inside TaskGroup serializer,
        # so it will not have its own endpoint, that's why
        # we need to explicitly define fields to exclude 'pulp_href' field.
        fields = ("message", "code", "total", "done", "suffix")



from unittest import TestCase

from rest_framework import serializers

from pulpcore.app.serializers import validate_unknown_fields


class TestValidateUnknownFields(TestCase):
    def test_unknown_field(self):
        """
        Test disjoint sets of `initial_data` with a single field and `defined_fields`.
        """
        initial_data = {"unknown1": "unknown"}
        defined_fields = {"field1": 1, "field2": 2}

        self.assertRaises(
            serializers.ValidationError, validate_unknown_fields, initial_data, defined_fields
        )

    def test_unknown_fields(self):
        """
        Test disjoint sets of `initial_data` with multiple fields and `defined_fields`.
        """
        initial_data = {"unknown1": "unknown", "unknown2": "unknown", "unknown3": "unknown"}
        defined_fields = {"field1": 1, "field2": 2}

        self.assertRaises(
            serializers.ValidationError, validate_unknown_fields, initial_data, defined_fields
        )

    def test_mixed_initial_data(self):
        """
        Test where `defined_fields` is a proper subset of the `initial_data`.
        """
        initial_data = {
            "field1": 1,
            "field2": 2,
            "unknown1": "unknown",
            "unknown2": "unknown",
            "unknown3": "unknown",
        }
        defined_fields = {"field1": 1, "field2": 2}
        self.assertRaises(
            serializers.ValidationError, validate_unknown_fields, initial_data, defined_fields
        )

    def test_mixed_incomplete_initial_data(self):
        """
        Test where `initial_data` and `defined_fields` are intersecting sets.
        """
        initial_data = {
            "field2": 2,
            "unknown1": "unknown",
            "unknown2": "unknown",
            "unknown3": "unknown",
        }
        defined_fields = {"field1": 1, "field2": 2}
        self.assertRaises(
            serializers.ValidationError, validate_unknown_fields, initial_data, defined_fields
        )

    def test_empty_defined_fields(self):
        """
        Test an empty `defined_fields`.
        """
        initial_data = {
            "field2": 2,
            "unknown1": "unknown",
            "unknown2": "unknown",
            "unknown3": "unknown",
        }
        defined_fields = {}
        self.assertRaises(
            serializers.ValidationError, validate_unknown_fields, initial_data, defined_fields
        )

    def test_validate_no_unknown_fields(self):
        """
        Test where the `initial_data` is equal to `defined_fields`.
        """
        initial_data = {"field1": 1, "field2": 2}
        defined_fields = {"field1": 1, "field2": 2}
        try:
            validate_unknown_fields(initial_data, defined_fields)
        except serializers.ValidationError:
            self.fail("validate_unknown_fields() failed improperly.")

    def test_validate_no_unknown_fields_no_side_effects(self):
        """
        Test validation success where the `initial_data` is equal to `defined_fields`
        and that `initial_data` and `defined_fields` are not mutated.
        """
        initial_data = {"field1": 1, "field2": 2}
        defined_fields = {"field1": 1, "field2": 2}
        try:
            validate_unknown_fields(initial_data, defined_fields)
        except serializers.ValidationError:
            self.fail("validate_unknown_fields() failed improperly.")

        self.assertEqual(initial_data, {"field1": 1, "field2": 2})
        self.assertEqual(defined_fields, {"field1": 1, "field2": 2})

    def test_ignored_fields_no_side_effects(self):
        """
        Test ignored fields in initial data don't cause side effects
        """
        # there's just the `csrfmiddlewaretoken` in the ignored_fields
        initial_data = {"field1": 1, "csrfmiddlewaretoken": 2}
        defined_fields = {"field1": 1}
        try:
            validate_unknown_fields(initial_data, defined_fields)
        except serializers.ValidationError:
            self.fail("validate_unknown_fields() failed improperly.")



"""Utilities for Pulpcore tests."""
from functools import partial
from unittest import SkipTest

from pulp_smash import config, selectors
from pulp_smash.pulp3.utils import require_pulp_3, require_pulp_plugins

from pulpcore.client.pulpcore import ApiClient


def set_up_module():
    """Skip tests Pulp 3 isn't under test or if pulpcore isn't installed."""
    require_pulp_3(SkipTest)
    require_pulp_plugins({"core"}, SkipTest)


skip_if = partial(selectors.skip_if, exc=SkipTest)  # pylint:disable=invalid-name
"""The ``@skip_if`` decorator, customized for unittest.

:func:`pulp_smash.selectors.skip_if` is test runner agnostic. This function is
identical, except that ``exc`` has been set to ``unittest.SkipTest``.
"""

cfg = config.get_config()
configuration = cfg.get_bindings_config()
core_client = ApiClient(configuration)



from pulpcore.app.models import Group
from pulpcore.app.serializers import RepositorySyncURLSerializer


# Model checks


def has_model_perms(request, view, action, permission):
    """
    Checks if the current user has a model-level permission.

    This is usable as a conditional check in an AccessPolicy. Here is an example checking for
    "file.add_fileremote" permission at the model-level.

    ::

        {
            ...
            "condition": "has_model_perms:file.add_fileremote",
        }

    Args:
        request (rest_framework.request.Request): The request being made.
        view (subclass rest_framework.viewsets.GenericViewSet): The view being checked for
            authorization.
        action (str): The action being performed, e.g. "destroy".
        permission (str): The name of the Permission to be checked. In the form
            `app_label.codename`, e.g. "core.delete_task".

    Returns:
        True if the user has the Permission named by the ``permission`` argument at the model-level.
            False otherwise.
    """
    return request.user.has_perm(permission)


def has_obj_perms(request, view, action, permission):
    """
    Checks if the current user has object-level permission on the specific object.

    The object in this case is the one the action is operating on, e.g. the URL
    ``/pulp/api/v3/tasks/15939b47-6b6d-4613-a441-939ca4ba6e63/`` is operating on the Task object
    with ``pk=15939b47-6b6d-4613-a441-939ca4ba6e63``.

    This is usable as a conditional check in an AccessPolicy. Here is an example checking for the
    "file.view_fileremote" permissions at the object-level.

    ::

        {
            ...
            "condition": "has_obj_perms:file.view_fileremote",
        }

    Args:
        request (rest_framework.request.Request): The request being made.
        view (subclass rest_framework.viewsets.GenericViewSet): The view being checked for
            authorization.
        action (str): The action being performed, e.g. "destroy".
        permission (str): The name of the Permission to be checked. In the form
            `app_label.codename`, e.g. "core.delete_task".

    Returns:
        True if the user has the Permission named by the ``permission`` argument on the object being
            operated on at the object-level. False otherwise.
    """
    obj = view.get_object()
    return request.user.has_perm(permission, obj)


def has_model_or_obj_perms(request, view, action, permission):
    """
    Checks if the current user has either model-level or object-level permissions.

    The object in this case is the one the action is operating on, e.g. the URL
    ``/pulp/api/v3/tasks/15939b47-6b6d-4613-a441-939ca4ba6e63/`` is operating on the Task object
    with ``pk=15939b47-6b6d-4613-a441-939ca4ba6e63``.

    This is usable as a conditional check in an AccessPolicy. Here is an example checking for
    "file.view_fileremote" permission at either the model-level or object-level.

    ::

        {
            ...
            "condition": "has_model_or_obj_perms:file.view_fileremote",
        }

    Args:
        request (rest_framework.request.Request): The request being made.
        view (subclass rest_framework.viewsets.GenericViewSet): The view being checked for
            authorization.
        action (str): The action being performed, e.g. "destroy".
        permission (str): The name of the Permission to be checked. In the form
            `app_label.codename`, e.g. "core.delete_task".

    Returns:
        True if the user has the Permission named by the ``permission`` argument at the model-level
            or on the object being operated on at the object-level. False otherwise.
    """
    if has_model_perms(request, view, action, permission):
        return True
    if has_obj_perms(request, view, action, permission):
        return True
    return False


# 'Remote' parameter checks


def has_remote_param_obj_perms(request, view, action, permission):
    """
    Checks if the current user has object-level permission on the ``remote`` object.

    The object in this case is the one specified by the ``remote`` parameter. For example when
    syncing the ``remote`` parameter is passed in as an argument.

    This is usable as a conditional check in an AccessPolicy. Here is an example checking for the
    "file.view_fileremote" permissions at the object-level.

    ::

        {
            ...
            "condition": "has_remote_param_obj_perms:file.view_fileremote",
        }

    Since it is checking a ``remote`` object the permission argument should be one of the following:

    * "file.change_fileremote" - Permission to change the ``FileRemote``.
    * "file.view_fileremote" - Permission to view the ``FileRemote``.
    * "file.delete_fileremote" - Permission to delete the ``FileRemote``.

    Args:
        request (rest_framework.request.Request): The request being made.
        view (subclass rest_framework.viewsets.GenericViewSet): The view being checked for
            authorization.
        action (str): The action being performed, e.g. "destroy".
        permission (str): The name of the Permission to be checked. In the form
            `app_label.codename`, e.g. "core.delete_task".

    Returns:
        True if the user has the Permission named by the ``permission`` argument on the ``remote``
            parameter at the object-level. False otherwise.
    """
    serializer = RepositorySyncURLSerializer(data=request.data, context={"request": request})
    serializer.is_valid(raise_exception=True)
    remote = serializer.validated_data.get("remote")
    return request.user.has_perm(permission, remote)


def has_remote_param_model_or_obj_perms(request, view, action, permission):
    """
    Checks if the current user has either model-level or object-level permissions on the ``remote``.

    The object in this case is the one specified by the ``remote`` parameter. For example when
    syncing the ``remote`` parameter is passed in as an argument.

    This is usable as a conditional check in an AccessPolicy. Here is an example checking for
    "file.view_fileremote" permission at either the model-level or object-level.

    ::

        {
            ...
            "condition": "has_remote_param_model_or_obj_perms:file.view_fileremote",
        }

    Since it is checking a ``remote`` object the permission argument should be one of the following:

    * "file.change_fileremote" - Permission to change the ``FileRemote``.
    * "file.view_fileremote" - Permission to view the ``FileRemote``.
    * "file.delete_fileremote" - Permission to delete the ``FileRemote``.

    Args:
        request (rest_framework.request.Request): The request being made.
        view (subclass rest_framework.viewsets.GenericViewSet): The view being checked for
            authorization.
        action (str): The action being performed, e.g. "destroy".
        permission (str): The name of the Permission to be checked. In the form
            `app_label.codename`, e.g. "core.delete_task".

    Returns:
        True if the user has the Permission named by the ``permission`` at the model-level or on the
            argument on the ``remote`` parameter at the object-level. False otherwise.
    """
    if has_model_perms(request, view, action, permission):
        return True
    if has_remote_param_obj_perms(request, view, action, permission):
        return True
    return False


# 'Repository' attribute checks


def has_repo_attr_obj_perms(request, view, action, permission):
    """
    Checks if the current user has object-level permission on a ``repository`` attribute.

    The object in this case is the one specified by the ``repository`` attribute of a resource
    which is being operated on. For example, when deleting a repository version, a ``repository``
    is its attribute.

    This is usable as a conditional check in an AccessPolicy. Here is an example checking for the
    "file.view_filerepository" permissions at the object-level.

    ::

        {
            ...
            "condition": "has_repo_attr_obj_perms:file.view_filerepository",
        }

    Since it is checking a ``repository`` object the permission argument should be one of the
    following:

    * "file.change_filerepository" - Permission to change the ``FileRepository``.
    * "file.view_filerepository" - Permission to view the ``FileRepository``.
    * "file.delete_filerepository" - Permission to delete the ``FileRepository``.
    * any custom permission a plugin has defined for their repository.

    Args:
        request (rest_framework.request.Request): The request being made.
        view (subclass rest_framework.viewsets.GenericViewSet): The view being checked for
            authorization.
        action (str): The action being performed, e.g. "destroy".
        permission (str): The name of the repository Permission to be checked. In the form
            `app_label.codename`, e.g. "file.delete_filerepository".

    Returns:
        True if the user has the Permission named by the ``permission`` argument on the
        ``repository`` attribute at the object-level. False otherwise.
    """
    plugin_repository = view.get_object().repository.cast()
    return request.user.has_perm(permission, plugin_repository)


def has_repo_attr_model_or_obj_perms(request, view, action, permission):
    """
    Checks if the current user has model-level or object-level permissions on a ``repository``.

    The object in this case is the one specified by the ``repository`` attribute of a resource
    which is being operated on. For example, when deleting a repository version, a ``repository``
    is its attribute.

    This is usable as a conditional check in an AccessPolicy. Here is an example checking for the
    "file.view_filerepository" permissions at the object-level.

    ::

        {
            ...
            "condition": "has_repo_attr_model_or_obj_perms:file.view_filerepository",
        }

    Since it is checking a ``repository`` object the permission argument should be one of the
    following:

    * "file.change_filerepository" - Permission to change the ``FileRepository``.
    * "file.view_filerepository" - Permission to view the ``FileRepository``.
    * "file.delete_filerepository" - Permission to delete the ``FileRepository``.
    * any custom permission a plugin has defined for their repository.

    Args:
        request (rest_framework.request.Request): The request being made.
        view (subclass rest_framework.viewsets.GenericViewSet): The view being checked for
            authorization.
        action (str): The action being performed, e.g. "destroy".
        permission (str): The name of the repository Permission to be checked. In the form
            `app_label.codename`, e.g. "file.delete_filerepository".

    Returns:
        True if the user has the Permission on the ``repository`` attribute named by the
        ``permission`` at the model or object level. False otherwise.
    """
    if has_model_perms(request, view, action, permission):
        return True
    if has_repo_attr_obj_perms(request, view, action, permission):
        return True
    return False


# `Group` permission checks


def has_group_obj_perms(request, view, action, permission):
    """
    Checks whether a user has the requested object permission on the Group in the URL.

    This check is meant to be used for relations nested beneath the group endpoint, e.g. the list
    of users to belong to that group. It will fail for other endpoints.

    ::

        {
            ...
            "condition": "has_group_obj_perms:core.group_delete",
        },

    Args:
        request (rest_framework.request.Request): The request being made.
        view (subclass rest_framework.viewsets.GenericViewSet): The view being checked for
            authorization.
        action (str): The action being performed, e.g. "destroy".
        permission (str): The name of the Group Permission to be checked. In the form
            `app_label.codename`, e.g. "core.group_change".

    Returns:
        True if the user has the Permission on the ``Group`` specified in the URL named by the
        ``permission`` at object level. False otherwise.
    """
    group_pk = request.resolver_match.kwargs["group_pk"]
    return request.user.has_perm(permission, Group.objects.get(pk=group_pk))


def has_group_model_or_obj_perms(request, view, action, permission):
    """
    Checks whether a user has the requested object or model permission on the Group in the URL.

    This check is meant to be used for relations nested beneath the group endpoint, e.g. the list
    of users to belong to that group. It will fail for other endpoints.

    ::

        {
            ...
            "condition": "has_group_model_or_obj_perms:core.group_delete",
        },

    Args:
        request (rest_framework.request.Request): The request being made.
        view (subclass rest_framework.viewsets.GenericViewSet): The view being checked for
            authorization.
        action (str): The action being performed, e.g. "destroy".
        permission (str): The name of the Group Permission to be checked. In the form
            `app_label.codename`, e.g. "core.group_change".

    Returns:
        True if the user has the Permission on the ``Group`` specified in the URL named by the
        ``permission`` at the model or object level. False otherwise.
    """
    return request.user.has_perm(permission) or has_group_obj_perms(
        request, view, action, permission
    )



import http.client
from gettext import gettext as _


class PulpException(Exception):
    """
    Base exception class for Pulp.
    """

    http_status_code = http.client.INTERNAL_SERVER_ERROR

    def __init__(self, error_code):
        """
        :param error_code: unique error code
        :type error_code: str
        """
        if not isinstance(error_code, str):
            raise TypeError(_("Error code must be an instance of str."))
        self.error_code = error_code

    def __str__(self):
        """
        Returns the string representation of the exception.

        Each concrete class that inherits from :class: `pulpcore.server.exception.PulpException` is
        expected to implement it's own __str__() method. The return value is used by Pulp when
        recording the exception in the database.
        """
        raise NotImplementedError("Subclasses of PulpException must implement a __str__() method")


def exception_to_dict(exc, traceback=None):
    """
    Return a dictionary representation of an Exception.

    :param exc: Exception that is being serialized
    :type exc: Exception
    :param traceback: String representation of a traceback generated when the exception occurred.
    :type traceback: str

    :return: dictionary representing the Exception
    :rtype: dict
    """
    return {"description": str(exc), "traceback": traceback}


class ResourceImmutableError(PulpException):
    """
    Exceptions that are raised due to trying to update an immutable resource
    """

    def __init__(self, model):
        """
        Args:
            model (pulpcore.app.models.Model): that the user is trying to update
        """
        super().__init__("PLP0003")
        self.model = model

    def __str__(self):
        msg = _("Cannot update immutable resource {model_pk} of type {model_type}").format(
            resource=str(self.model.pk), type=type(self.model).__name__
        )
        return msg


class AdvisoryLockError(Exception):
    """Exception to signal that a lock could not be acquired."""


class TimeoutException(PulpException):
    """
    Exception to signal timeout error.
    """

    def __init__(self, url):
        """
        :param url: the url the download for timed out
        :type url: str
        """
        super().__init__("PLP0005")
        self.url = url

    def __str__(self):
        return _(
            "Request timed out for {}. Increasing the total_timeout value on the remote might help."
        ).format(self.url)



# Generated by Django 2.2.6 on 2019-11-12 14:26

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0014_remove_repository_plugin_managed'),
    ]

    operations = [
        migrations.RenameField(
            model_name='remote',
            old_name='ssl_ca_certificate',
            new_name='ca_cert',
        ),
        migrations.RenameField(
            model_name='remote',
            old_name='ssl_client_certificate',
            new_name='client_cert',
        ),
        migrations.RenameField(
            model_name='remote',
            old_name='ssl_client_key',
            new_name='client_key',
        ),
        migrations.RenameField(
            model_name='remote',
            old_name='ssl_validation',
            new_name='tls_validation',
        ),
    ]
